# Copyright (c) 2023 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of qonnx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import clize
import numpy as np
import os
import urllib.request

import qonnx.core.onnx_exec as oxe
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.util.cleanup import cleanup

# utility functions to fetch models and data for
# testing various qonnx transformations

a2q_rn18_preproc_mean = np.asarray([0.491, 0.482, 0.447], dtype=np.float32)
a2q_rn18_preproc_std = np.asarray([0.247, 0.243, 0.262], dtype=np.float32)
a2q_rn18_int_range = (0, 255)
a2q_rn18_iscale = 1 / 255
a2q_rn18_rmin = (a2q_rn18_int_range[0] * a2q_rn18_iscale - a2q_rn18_preproc_mean) / a2q_rn18_preproc_std
a2q_rn18_rmax = (a2q_rn18_int_range[1] * a2q_rn18_iscale - a2q_rn18_preproc_mean) / a2q_rn18_preproc_std
a2q_rn18_scale = (1 / a2q_rn18_preproc_std) * a2q_rn18_iscale
a2q_rn18_bias = -a2q_rn18_preproc_mean * a2q_rn18_preproc_std
a2q_rn18_common = {
    "input_shape": (1, 3, 32, 32),
    "input_range": (a2q_rn18_rmin, a2q_rn18_rmax),
    "int_range": a2q_rn18_int_range,
    "scale": a2q_rn18_scale,
    "bias": a2q_rn18_bias,
}
a2q_rn18_urlbase = "https://github.com/fastmachinelearning/qonnx_model_zoo/releases/download/a2q-20240905/"

a2q_model_details = {
    "rn18_w4a4_a2q_16b": {
        "description": "4-bit ResNet-18 on CIFAR-10, A2Q 16-bit accumulators",
        "url": a2q_rn18_urlbase + "quant_resnet18_w4a4_a2q_16b-d4bfa990.onnx",
        **a2q_rn18_common,
    },
    "rn18_w4a4_a2q_15b": {
        "description": "4-bit ResNet-18 on CIFAR-10, A2Q 15-bit accumulators",
        "url": a2q_rn18_urlbase + "quant_resnet18_w4a4_a2q_15b-eeca8ac2.onnx",
        **a2q_rn18_common,
    },
    "rn18_w4a4_a2q_14b": {
        "description": "4-bit ResNet-18 on CIFAR-10, A2Q 14-bit accumulators",
        "url": a2q_rn18_urlbase + "quant_resnet18_w4a4_a2q_14b-563cf426.onnx",
        **a2q_rn18_common,
    },
    "rn18_w4a4_a2q_13b": {
        "description": "4-bit ResNet-18 on CIFAR-10, A2Q 13-bit accumulators",
        "url": a2q_rn18_urlbase + "quant_resnet18_w4a4_a2q_13b-d3cae293.onnx",
        **a2q_rn18_common,
    },
    "rn18_w4a4_a2q_12b": {
        "description": "4-bit ResNet-18 on CIFAR-10, A2Q 12-bit accumulators",
        "url": a2q_rn18_urlbase + "quant_resnet18_w4a4_a2q_12b-fb3a0f8a.onnx",
        **a2q_rn18_common,
    },
    "rn18_w4a4_a2q_plus_16b": {
        "description": "4-bit ResNet-18 on CIFAR-10, A2Q+ 16-bit accumulators",
        "url": a2q_rn18_urlbase + "quant_resnet18_w4a4_a2q_plus_16b-09e47feb.onnx",
        **a2q_rn18_common,
    },
    "rn18_w4a4_a2q_plus_15b": {
        "description": "4-bit ResNet-18 on CIFAR-10, A2Q+ 15-bit accumulators",
        "url": a2q_rn18_urlbase + "quant_resnet18_w4a4_a2q_plus_15b-10e7bc83.onnx",
        **a2q_rn18_common,
    },
    "rn18_w4a4_a2q_plus_14b": {
        "description": "4-bit ResNet-18 on CIFAR-10, A2Q+ 14-bit accumulators",
        "url": a2q_rn18_urlbase + "quant_resnet18_w4a4_a2q_plus_14b-8db8c78c.onnx",
        **a2q_rn18_common,
    },
    "rn18_w4a4_a2q_plus_13b": {
        "description": "4-bit ResNet-18 on CIFAR-10, A2Q+ 13-bit accumulators",
        "url": a2q_rn18_urlbase + "quant_resnet18_w4a4_a2q_plus_13b-f57b05ce.onnx",
        **a2q_rn18_common,
    },
    "rn18_w4a4_a2q_plus_12b": {
        "description": "4-bit ResNet-18 on CIFAR-10, A2Q+ 12-bit accumulators",
        "url": a2q_rn18_urlbase + "quant_resnet18_w4a4_a2q_plus_12b-1e2aca29.onnx",
        **a2q_rn18_common,
    },
}

test_model_details = {
    "FINN-CNV_W2A2": {
        "description": "2-bit VGG-10-like CNN on CIFAR-10",
        "url": (
            "https://raw.githubusercontent.com/fastmachinelearning/"
            "QONNX_model_zoo/main/models/CIFAR10/Brevitas_FINN_CNV/CNV_2W2A.onnx"
        ),
        "input_shape": (1, 3, 32, 32),
        "input_range": (0, +1),
    },
    "FINN-CNV_W1A2": {
        "description": "1/2-bit VGG-10-like CNN on CIFAR-10",
        "url": (
            "https://raw.githubusercontent.com/fastmachinelearning/"
            "QONNX_model_zoo/main/models/CIFAR10/Brevitas_FINN_CNV/CNV_1W2A.onnx"
        ),
        "input_shape": (1, 3, 32, 32),
        "input_range": (0, +1),
    },
    "FINN-CNV_W1A1": {
        "description": "1-bit VGG-10-like CNN on CIFAR-10",
        "url": (
            "https://raw.githubusercontent.com/fastmachinelearning/"
            "QONNX_model_zoo/main/models/CIFAR10/Brevitas_FINN_CNV/CNV_1W1A.onnx"
        ),
        "input_shape": (1, 3, 32, 32),
        "input_range": (0, +1),
    },
    "FINN-TFC_W1A1": {
        "description": "1-bit tiny MLP on MNIST",
        "url": (
            "https://github.com/fastmachinelearning/QONNX_model_zoo/"
            "raw/main/models/MNIST/Brevitas_FINN_TFC/TFC/TFC_1W1A.onnx"
        ),
        "input_shape": (1, 1, 28, 28),
        "input_range": (0, +1),
    },
    "FINN-TFC_W1A2": {
        "description": "1/2-bit tiny MLP on MNIST",
        "url": (
            "https://github.com/fastmachinelearning/QONNX_model_zoo/"
            "raw/main/models/MNIST/Brevitas_FINN_TFC/TFC/TFC_1W2A.onnx"
        ),
        "input_shape": (1, 1, 28, 28),
        "input_range": (0, +1),
    },
    "FINN-TFC_W2A2": {
        "description": "2-bit tiny MLP on MNIST",
        "url": (
            "https://github.com/fastmachinelearning/QONNX_model_zoo/"
            "raw/main/models/MNIST/Brevitas_FINN_TFC/TFC/TFC_2W2A.onnx"
        ),
        "input_shape": (1, 1, 28, 28),
        "input_range": (0, +1),
    },
    "RadioML_VGG10": {
        "description": "8-bit VGG-10-like CNN on RadioML 2018",
        "url": (
            "https://github.com/Xilinx/brevitas-radioml-challenge-21/raw/"
            "9eef6a2417d6a0c078bfcc3a4dc95033739c5550/sandbox/notebooks/models/pretrained_VGG10_w8a8_20_export.onnx"
        ),
        "input_shape": (1, 2, 1024),
        "input_range": (-1, +1),
    },
    "Conv_bias_example": {
        "description": "",
        "url": "https://zenodo.org/record/7626922/files/super_resolution.onnx",
        "input_shape": (1, 1, 28, 28),
        "input_range": (-1, +1),
    },
    "MobileNetv1-w4a4": {
        "description": "4-bit MobileNet-v1 on ImageNet",
        "url": (
            "https://raw.githubusercontent.com/fastmachinelearning/"
            "qonnx_model_zoo/main/models/ImageNet/Brevitas_FINN_mobilenet/mobilenet_4W4A.onnx"
        ),
        "input_shape": (1, 3, 224, 224),
        "input_range": (0, 1),
    },
    **a2q_model_details,
}


test_model_keys = clize.parameters.mapped(
    [(x, [x], test_model_details[x]["description"]) for x in test_model_details.keys()]
)


def download_model(test_model: test_model_keys, *, dl_dir="/tmp", do_cleanup=False, return_modelwrapper=False):
    qonnx_url = test_model_details[test_model]["url"]
    # download test data
    dl_file = dl_dir + f"/{test_model}.onnx"
    ret = dl_file
    if not os.path.isfile(dl_file):
        urllib.request.urlretrieve(qonnx_url, dl_file)
    if do_cleanup:
        # run cleanup with default settings
        out_file = dl_dir + f"/{test_model}_clean.onnx"
        cleanup(dl_file, out_file=out_file)
        ret = out_file
    if return_modelwrapper:
        ret = ModelWrapper(ret)
    return ret


def qonnx_download_model():
    clize.run(download_model)


def get_random_input(test_model, seed=42):
    rng = np.random.RandomState(seed)
    input_shape = test_model_details[test_model]["input_shape"]
    (low, high) = test_model_details[test_model]["input_range"]
    # some models spec per-channel ranges, be conservative for those
    if isinstance(low, np.ndarray):
        low = low.max()
    if isinstance(high, np.ndarray):
        high = high.min()
    size = np.prod(np.asarray(input_shape))
    input_tensor = rng.uniform(low=low, high=high, size=size)
    input_tensor = input_tensor.astype(np.float32)
    input_tensor = input_tensor.reshape(input_shape)
    return input_tensor


def get_golden_in_and_output(test_model, seed=42):
    model = download_model(test_model, do_cleanup=True, return_modelwrapper=True)
    input_tensor = get_random_input(test_model, seed=seed)
    input_dict = {model.graph.input[0].name: input_tensor}
    golden_output_dict = oxe.execute_onnx(model, input_dict)
    golden_result = golden_output_dict[model.graph.output[0].name]
    return input_tensor, golden_result



# Set pytest parameters
import pytest

# Numpy for handling simulation of tensor operations
import numpy as np

# Helper for creating ONNX nodes
from onnx import NodeProto, TensorProto  # noqa
from onnx import helper as oh  # noqa

# QONNX wrapper of ONNX model graphs
from qonnx.core.modelwrapper import ModelWrapper  # noqa

# Execute QONNX model graphs
from qonnx.core.onnx_exec import execute_onnx  # noqa

# QONNX quantizer function modeling the behavior of the Quant operator
from qonnx.custom_op.general.quant import quant as quant_fn  # noqa

# QONNX graph transformations for inferring datatypes and shapes required by
# test setup
from qonnx.transformation.infer_datatypes import InferDataTypes  # noqa
from qonnx.transformation.infer_shapes import InferShapes  # noqa

# Graph transformation to be tested: Transposes the initializers to Quantizer if
# ALL inputs are initializers
from qonnx.transformation.quant_constant_folding import FoldTransposeIntoQuantInit  # noqa

# QONNX utility  for creating models from ONNX graphs
from qonnx.util.basic import qonnx_make_model  # noqa


@pytest.mark.parametrize("quant_init", [True, False])
@pytest.mark.parametrize("signed", [0, 1])
@pytest.mark.parametrize("narrow", [0, 1])
@pytest.mark.parametrize("rounding_mode", ["ROUND"])
@pytest.mark.parametrize("shape", [(16, 8, 12)])
@pytest.mark.parametrize(
    "perm",
    [
        # All axis permutations
        (0, 1, 2),
        (0, 2, 1),
        (1, 0, 2),
        (1, 2, 0),
        (2, 0, 1),
        (2, 1, 0),
    ],
)
@pytest.mark.parametrize("scale", [0.01])
@pytest.mark.parametrize("zeropoint", [0])
@pytest.mark.parametrize("bitwidth", [8])
# Tests the FoldTransposeIntoQuantInit transformation
def test_fold_transpose_into_quant_init(quant_init, signed, narrow, rounding_mode, shape, perm, scale, zeropoint, bitwidth):
    # Prepare the quantizer node attributes and input/output lists
    node_attrs = {
        # Type of the operation
        "op_type": "Quant",
        # This operator type is defined within QONNX
        "domain": "qonnx.custom_op.general",
        # List the inputs to the operator in order
        # Note: The proper input followed by initializers configuring the
        # quantizer
        "inputs": ["input", "scale", "zeropoint", "bitwidth"],
        # List the outputs of the operator in order
        # Note: Intermediate feeds to the next operator input
        "outputs": ["intermediate"],
        # Whether the quantization interval should be signed or not
        # (e.g. at 8b unsigned=[0, 255] vs signed=[-128, 127])
        "signed": signed,
        # When signed=1, whether to use narrow range or not
        # (e.g. at 8b regular=[-128, 127] vs narrow=[-127, 127])
        "narrow": narrow,
        # The rounding mode, which is used for the quant function
        "rounding_mode": rounding_mode,
    }
    # Create a dummy quantizer node
    quant = oh.make_node(**node_attrs, name="Quant")
    # Attach a Transpose operation to the quantizer
    transpose = oh.make_node("Transpose", ["intermediate"], ["output"], name="Transpose", perm=perm)
    # Derive the transposed shape
    transposed_shape = np.transpose(np.zeros(shape), perm).shape
    # Create tensor information for the input, intermediate and output tensor
    x = oh.make_tensor_value_info("input", TensorProto.FLOAT, shape)  # noqa
    y = oh.make_tensor_value_info("output", TensorProto.FLOAT, transposed_shape)
    # Create the initializer tensors for quantizer parameters
    s = oh.make_tensor_value_info("scale", TensorProto.FLOAT, (1,))
    z = oh.make_tensor_value_info("zeropoint", TensorProto.FLOAT, (1,))
    b = oh.make_tensor_value_info("bitwidth", TensorProto.FLOAT, (1,))
    # Create the graph connecting the nodes and tensors
    graph = oh.make_graph(
        [quant, transpose],
        "quant-transpose",
        [x, s, z, b],
        [y],
    )
    # Wrap the graph in an QONNX model wrapper
    model = ModelWrapper(qonnx_make_model(graph, producer_name="qonnx-tests"))
    # Add the actual initializers to the initializer tensors
    model.set_initializer("scale", np.array(scale))
    model.set_initializer("zeropoint", np.array(zeropoint))
    model.set_initializer("bitwidth", np.array(bitwidth))
    # Prepare the model graph by inferring all missing shape and datatype
    # information
    model = model.transform(InferShapes())
    model = model.transform(InferDataTypes())

    # Get a random dummy input for testing
    x = np.random.rand(*shape)  # noqa
    # Fill the execution context with dummy input data
    context = {"input": x}

    # Some test cases even turn the input into an initializer
    if quant_init:
        # Turn the model input into an initializer
        model.set_initializer("input", x)
        # Clear the execution context removing the input as it is now baked into
        # the model graph
        context = {}

    # Run the transformation to be tested
    model = model.transform(FoldTransposeIntoQuantInit())
    # Verify that shape and datatype inference still works
    # Note: This has been an issue, please see
    #   https://github.com/fastmachinelearning/qonnx/issues/77
    model = model.transform(InferShapes())
    model = model.transform(InferDataTypes())

    # For the case of quant-initializers there must not be a Transpose left
    # after transforming and contrariwise, the Transpose must remain in place if
    # there is non-initializer input.
    assert quant_init != ("Transpose" in [n.op_type for n in model.graph.node])

    # Execute the ONNX model
    o_produced = execute_onnx(model, context)["output"]
    # Use numpy and QONNX quantizer to generate expectation
    o_expected = np.transpose(
        quant_fn(x, np.array(scale), np.array(zeropoint), np.array(bitwidth), signed, narrow, rounding_mode), perm
    )

    # The output must match the "manual" execution using numpy
    assert np.allclose(o_produced, o_expected)



# Copyright (c) 2024 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of qonnx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import numpy as np
import onnx
from onnx import TensorProto

from qonnx.transformation.base import Transformation
from qonnx.transformation.general import SortGraph
from qonnx.transformation.infer_shapes import InferShapes
from qonnx.util.cleanup import cleanup_model


def create_quantnode(
    model,
    quantnode_input,
    quantnode_output_shape,
    scale_value,
    zeropoint_value,
    bitwidth_value,
    narrow,
    signed,
    rounding_mode,
):
    quant_tensor = onnx.helper.make_tensor_value_info(
        model.make_new_valueinfo_name(), TensorProto.FLOAT, quantnode_output_shape
    )
    model.graph.value_info.append(quant_tensor)

    scale_tensor = np.array(scale_value).astype(np.float32)
    s_value = onnx.helper.make_tensor_value_info(model.make_new_valueinfo_name(), TensorProto.FLOAT, quantnode_output_shape)
    model.graph.value_info.append(s_value)
    model.set_initializer(s_value.name, scale_tensor)

    zeropt_tensor = np.array(zeropoint_value).astype(np.float32)
    z_value = onnx.helper.make_tensor_value_info(model.make_new_valueinfo_name(), TensorProto.FLOAT, quantnode_output_shape)
    model.graph.value_info.append(z_value)
    model.set_initializer(z_value.name, zeropt_tensor)

    bitwidth_tensor = np.array(bitwidth_value).astype(np.float32)
    b_value = onnx.helper.make_tensor_value_info(model.make_new_valueinfo_name(), TensorProto.FLOAT, [1])
    model.graph.value_info.append(b_value)
    model.set_initializer(b_value.name, bitwidth_tensor)

    quantnode = onnx.helper.make_node(
        "Quant",
        inputs=[quantnode_input, s_value.name, z_value.name, b_value.name],
        outputs=[quant_tensor.name],
        name="Quant_" + quantnode_input,
        narrow=narrow,
        signed=signed,
        rounding_mode=rounding_mode,
    )

    return quantnode, quant_tensor


def adjust_graph(model, input_positions, node_name, quantized_nodes):
    for pos in input_positions:
        node_details = (node_name, pos[0])
        if node_details not in quantized_nodes:  # not quantizing for same node_inp/out index.
            node_in_focus = model.get_node_from_name(node_name)

            if pos[0][0] == "input":
                quantnode_input = node_in_focus.input[pos[0][1]]
                consumer_node = node_in_focus
                producer_node = model.find_producer(quantnode_input)
                if producer_node is None or producer_node.op_type != "Quant":
                    quantization_to_perform = True
                else:
                    quantization_to_perform = False
            else:
                quantnode_input = node_in_focus.output[pos[0][1]]
                consumer_node = model.find_consumer(quantnode_input)
                producer_node = model.find_producer(quantnode_input)
                if consumer_node is None or consumer_node.op_type != "Quant":
                    quantization_to_perform = True
                else:
                    quantization_to_perform = False
            if quantization_to_perform is True:
                quantnode_output_shape = model.get_tensor_shape(quantnode_input)  # Step: 3
                quantnode, quant_tensor = create_quantnode(
                    model,
                    quantnode_input,
                    quantnode_output_shape,
                    scale_value=pos[1][0],
                    zeropoint_value=pos[1][1],
                    bitwidth_value=pos[1][2],
                    narrow=pos[1][3],
                    signed=pos[1][4],
                    rounding_mode=pos[1][5],
                )

                if consumer_node is not None:
                    node_pos = model.get_node_index(consumer_node)
                    model.graph.node[node_pos].input[pos[0][1]] = quant_tensor.name
                    model.graph.node.append(quantnode)
                else:
                    model.graph.value_info.remove(quant_tensor)
                    model.graph.node.append(quantnode)
                    model.graph.output.insert(0, quant_tensor)
                    model.graph.output.pop(1)

                model = model.transform(SortGraph())
                quantized_nodes.append(node_details)
            else:
                print(f"{pos[0][0]} index {pos[0][1]} of {node_name} is already quantized.")
        else:
            print(f"{pos[0][0]} index {pos[0][1]} of {node_name} is already quantized.")
            continue

    return model


class QuantizeGraph(Transformation):
    """This transformation can be used to introduce a Quant node for a specific type of node in the graph.
    Users would be able to specify the location of the quant node by providing the input and output index
    as the parameters.

        1) Expectations:
            a) Onnx model in the modelwraper format.
            b) Model must be cleaned using qonnx.util.cleanup.cleanup_model()
            c) Batchsize to be set.

        2) Steps to transform are:
            Step1: Finding the input for the quant node.
            Step2: Finding the consumer of the quant node output.
            Step3: Finding the shape for the output tensor of quant node.
            Note: The output tensor of the quant node must have the same shape as the consumer of the input
                    to the quant node.

        3) Input:
            A dict "quantnode_map" specifying the criterion, positions, and input parameters like
            scale, bitwidth, zeropoint, and others for a specific quantnode.

            Criterion:
                a) name: This will allow users to add quant nodes for specific node like "Conv_0" and "Gemm_0".
                        Note: using this users can have quant nodes with different parameters. Ex: quantizing
                        "Conv_0" and "Conv_1"  with bitwidth of 4 and 6, respectively.
                b) op_type: This will allow users to add quant nodes for all nodes of a particular op_type such
                            as, "Conv", "Gemm", and others.
                            Note: All quant nodes created using op_type criterion will have the same input
                            parameters (scale, zeropoint, bitwidth, and others.)
                c) name and op_type: In this case, quant nodes will be added with precedence to "Name"
                                    in comparison to "op_type".

            Positions:  ("input", index) or  ("output", index)
                a) "input":  indicates that the user want to quantize the input of the selected node.
                b) "output": indicates that the user want to quantize the output of the selected node.
                c) index: refers to the input/output index to quantize (a node can have multiple inputs and outputs)

            Parameters (to quant node) are provided as (scale, zeropoint, bitwidth, narrow, signed, rounding_mode)

                a) Inputs: scale, zeropoint, bitwidth.
                b) Attributes: narrow, signed, rounding_mode.

        4) Assert:
                a) The input is a dictionary representing the node names as keys and a list of quant positions
                   as values.
                b) The input dictionary must have atleast one mac node (Conv, gemm, matmul) for the transformation.

        5) Return:
                Returns a model with new quant nodes created at the positions specified using the "quantnode_map".

        6) Example:
                quantnode_map = {"name": {"Conv_0": [(("input", 0), (1, 0, 8, 0, 1, "ROUND")),
                                                (("input", 1), (1, 0, 8, 0, 1, "ROUND")),
                                                (("output", 0), (1, 0, 8, 0, 1, "ROUND"))],
                                        "Conv_1": [(("input", 0), (1, 0, 8, 0, 1, "ROUND"))],
                                        "Conv_2": [(("input", 1), (1, 0, 8, 0, 1, "ROUND")),
                                                (("output", 0), (1, 0, 8, 0, 1, "ROUND"))]},

                                "op_type": {"Gemm": [(("input", 0), (1, 0, 8, 0, 1, "ROUND")),
                                                   (("input", 1), (1, 0, 8, 0, 1, "ROUND")),
                                                   (("input", 2), (1, 0, 8, 0, 1, "ROUND")),
                                                   (("output", 0), (1, 0, 8, 0, 1, "ROUND"))]}}
    """

    def __init__(self, quantnode_map):
        super().__init__()
        self.quantnode_map = quantnode_map

    def apply(self, model):
        model = model.transform(InferShapes())
        if type(self.quantnode_map) == dict:
            selection_type = self.quantnode_map.keys()
            if set(selection_type) <= {"name", "op_type"}:
                quantized_nodes = []
                if "name" in selection_type:
                    by_name = self.quantnode_map["name"]  # dict with unique names and list of positions.
                    node_list_by_name = by_name.keys()  # node names specified by the user for quant nodes.
                    for node_name in node_list_by_name:
                        input_positions = by_name[node_name]  # input positions to introduce quant nodes.
                        model = adjust_graph(model, input_positions, node_name, quantized_nodes)
                if "op_type" in selection_type:
                    by_op_type = self.quantnode_map["op_type"]  # dict with the unique names and list of positions.
                    op_list = by_op_type.keys()
                    for op in op_list:
                        node_list = model.get_nodes_by_op_type(op)  # List of all nodes with the operation type "op".
                        input_positions = by_op_type[op]
                        for node in node_list:
                            node_name = node.name
                            model = adjust_graph(model, input_positions, node_name, quantized_nodes)
                model = cleanup_model(model)
            else:
                raise Exception("Unsupported selection type")
        else:
            raise TypeError("Input must be a dictionary.")

        graph_modified = False

        return (model, graph_modified)



# Set pytest parameters
import pytest

# Numpy for handling simulation of tensor operations
import numpy as np

# Helper for creating ONNX nodes
from onnx import TensorProto
from onnx import helper as oh

# QONNX wrapper of ONNX model graphs
from qonnx.core.modelwrapper import ModelWrapper

# Execute QONNX model graphs
from qonnx.core.onnx_exec import execute_onnx

# Graph transformation to be tested: Sorts the input list of commutative
# operations to have all dynamic inputs first followed by all initializer inputs
from qonnx.transformation.general import SortCommutativeInputsInitializerLast

# QONNX utility  for creating models from ONNX graphs
from qonnx.util.basic import qonnx_make_model


# Specify how many inputs the test should cover
@pytest.mark.parametrize("num_inputs", [4, 5, 6])
# Specify which inputs should be turned into initializers
@pytest.mark.parametrize(
    # fmt: off
    "initializers", [[], [0], [1], [0, 1], [0, 3], [0, 1, 2, 3]]
    # fmt: on
)
# Tests the SortCommutativeInputsInitializerLast transformation
def test_sort_commutative_inputs_initializer_last(num_inputs, initializers):
    # Generate the input tensor names
    inputs = [f"in{i}" for i in range(num_inputs)]
    # We will use the Sum ONNX operation to test this behavior, as it allows for
    # arbitrary many inputs
    node = oh.make_node(
        # fmt: off
        op_type="Sum", inputs=inputs, outputs=["out"], name="Sum"
        # fmt: on
    )
    # Create value infos for all input and the output tensor
    inputs = [
        # fmt: off
        oh.make_tensor_value_info(i, TensorProto.FLOAT, (16,)) for i in inputs
        # fmt: on
    ]
    out = oh.make_tensor_value_info("out", TensorProto.FLOAT, (16,))
    # Make a graph comprising the Sum node and value infos for all inputs and
    # the output
    graph = oh.make_graph([node], inputs=inputs, outputs=[out], name="Sum")
    # Wrap the graph in an QONNX model wrapper
    model = ModelWrapper(qonnx_make_model(graph, producer_name="qonnx-tests"))
    # Prepare the execution context
    context = {f"in{i}": np.random.rand(16) for i in range(num_inputs)}
    # Make sure all inputs are of type float32
    context = {key: value.astype(np.float32) for key, value in context.items()}
    # Turn selected inputs into initializers
    for i in initializers:
        model.set_initializer(f"in{i}", context[f"in{i}"])

    # Execute the ONNX model before transforming
    out_expected = execute_onnx(model, context)["out"]
    # Apply the transformation to be tested
    # Note: No cleanup, as the tested transformation is part of the cleanup, and
    # we want to test this in isolation
    model = model.transform(
        # fmt: off
        SortCommutativeInputsInitializerLast(), cleanup=False
        # fmt: on
    )
    # Execute the ONNX model after transforming
    out_produced = execute_onnx(model, context)["out"]

    # Start with no initializer input seen so far
    seen_initializer = False
    # Verify that no "dynamic" input follows an initializer input
    for i in model.graph.node[0].input:
        # Keep track of when an initializer has been seen
        if model.get_initializer(i) is not None:
            seen_initializer = True
        # If there has already been an initializer, this input must be an
        # initializer as well
        assert (
            not seen_initializer or model.get_initializer(i) is not None
        ), "Non-initializer input following initializer after sorting"

    # Outputs before and after must match
    assert np.allclose(out_produced, out_expected)



# Copyright (c) 2023 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import clize
import numpy as np
import os
from matplotlib import pyplot as plt
from tqdm import tqdm

from qonnx.core.modelwrapper import ModelWrapper
from qonnx.util.range_analysis import (
    REPORT_MODE_RANGE,
    REPORT_MODE_STUCKCHANNEL,
    REPORT_MODE_ZEROSTUCKCHANNEL,
    report_mode_options,
    report_modes,
)


def update_tensor_stats(tensor, axes, ret_dict={}):
    shp = tensor.shape
    if ret_dict == {}:
        ret_dict["shape"] = shp
    else:
        assert ret_dict["shape"] == shp
    for axis in axes:
        tensor_new = np.moveaxis(tensor, axis, 0).reshape(shp[axis], -1)
        ret_axis = {
            "min": np.min(tensor_new, axis=1),
            "max": np.max(tensor_new, axis=1),
        }
        axis_name = "axis%d" % axis
        if axis_name in ret_dict:
            ret_dict[axis_name] = ret_axis
            ret_axis["min"] = np.minimum(ret_axis["min"], ret_dict[axis_name]["min"])
            ret_axis["max"] = np.maximum(ret_axis["max"], ret_dict[axis_name]["max"])
        ret_dict[axis_name] = ret_axis
    return ret_dict


def tensor_stats(
    modelwrapper_or_filename,
    act_dump_dir: str,
    output_stats_dir: str,
    *,
    axes="1",
    plot=True,
    report_mode: report_mode_options = REPORT_MODE_RANGE
):
    assert report_mode in report_modes, "Unknown report_mode"
    if not isinstance(modelwrapper_or_filename, ModelWrapper):
        model = ModelWrapper(modelwrapper_or_filename)
    else:
        model = modelwrapper_or_filename
    if not isinstance(axes, list):
        axes = [int(x.strip()) for x in axes.split(",")]
    if not os.path.isdir(output_stats_dir):
        os.makedirs(output_stats_dir)

    all_tensor_dump_files = []
    all_tensor_dump_files = [f for f in os.listdir(act_dump_dir) if os.path.isfile(os.path.join(act_dump_dir, f))]
    all_tensor_dump_files = [f for f in all_tensor_dump_files if f.endswith(".npy")]
    tensorwise_stats = {}
    stuck_chans = {}
    for outp in tqdm(model.graph.output, "Tensors"):
        tname = outp.name
        t_files = [f for f in all_tensor_dump_files if f.startswith(tname)]
        tensorwise_stats[tname] = {}
        for f in tqdm(t_files, "Batches"):
            t_file = np.load(os.path.join(act_dump_dir, f))
            tensorwise_stats[tname] = update_tensor_stats(t_file, axes=axes, ret_dict=tensorwise_stats[tname])
        tstats = tensorwise_stats[tname]
        for axis in axes:
            axis_name = "axis%d" % axis
            axis_data = tstats[axis_name]
            axis_min = axis_data["min"]
            axis_max = axis_data["max"]
            tensor_stuck_chans = np.nonzero(axis_min == axis_max)[0]
            if report_mode in [REPORT_MODE_STUCKCHANNEL, REPORT_MODE_ZEROSTUCKCHANNEL]:
                if len(tensor_stuck_chans) > 0:
                    list_stuck_chans = list(tensor_stuck_chans)
                    list_stuck_values = list(axis_min[tensor_stuck_chans])
                    stuck_chans[tname] = list(zip(list_stuck_chans, list_stuck_values))
        if plot:
            for axis in axes:
                axis_name = "axis%d" % axis
                data = tensorwise_stats[tname][axis_name]
                axis_min = data["min"]
                axis_max = data["max"]
                axis_range = axis_max - axis_min
                chans = [i for i in range(len(axis_min))]
                plt.clf()
                plt.figure(constrained_layout=True, figsize=(5, len(axis_min) / 3))
                bars = plt.barh(chans, axis_range, left=axis_min)
                bar_labels = [str((axis_min[i], axis_max[i])) for i in range(len(axis_min))]
                plt.bar_label(bars, bar_labels)
                plt.yticks([x for x in range(len(axis_range))])
                plt.xlabel("Channel number")
                plt.ylabel("Channel range")
                plt.title("Observed range for %s_%s" % (tname, axis_name))

                plt.savefig(output_stats_dir + "/%s_%s.png" % (tname, axis_name))

    if report_mode == REPORT_MODE_RANGE:
        ret = tensorwise_stats
    elif report_mode == REPORT_MODE_STUCKCHANNEL:
        ret = stuck_chans
    elif report_mode == REPORT_MODE_ZEROSTUCKCHANNEL:
        # only leave channels that are stuck at zero
        # value info removed since implicitly 0
        ret = stuck_chans
        new_ret = {}
        for tname, schans in ret.items():
            schans_only_zero = set([x[0] for x in schans if x[1] == 0])
            if len(schans_only_zero) > 0:
                new_ret[tname] = schans_only_zero
        ret = new_ret
    return ret


def main():
    clize.run(tensor_stats)


if __name__ == "__main__":
    main()



# Copyright (c) 2020 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
import warnings

from qonnx.transformation.base import Transformation
from qonnx.transformation.general import RemoveUnusedTensors
from qonnx.transformation.infer_shapes import InferShapes
from qonnx.util.basic import get_by_name

# FINN currently handles convolutions (when e.g lowering them
# to matmuls) with the assumption that they operate on 4D tensors
# shaped as (N,C,H,W). H/W can be 1 for convolutions on 1D data.
# This transformation converts a graph with 3D tensors to the expected
# 4D format. Note: the transformation only works for certain node types;
# see _find_invalid_nodes below.


def _find_invalid_nodes(model):
    """
    Check whether the graph contains any node types that are not supported by the
    Change3Dto4DTensors transformation.

    """
    valid_nodes = [
        "Add",
        "Mul",
        "BatchNormalization",
        "MultiThreshold",
        "Conv",
        "Transpose",
        "LogSoftmax",
        "ArgMax",
        "Div",
        "TopK",
        "MatMul",
        "Flatten",
        "Reshape",
        "MaxPool",
        "Upsample",
        "Resize",
    ]
    invalid_nodes = []
    for n in model.graph.node:
        node_op_type = n.op_type
        if node_op_type in valid_nodes:
            continue
        else:
            invalid_nodes.append(node_op_type)

    return invalid_nodes


class Change3DTo4DTensors(Transformation):
    """
    Replaces 3D tensors with 4D tensors assuming the following format:
    [N, C, H] -> [N, C, H, 1].
    The attributes of a (specific) set of supported nodes are changed accordingly.
    If the graph contains unsupported nodes, a warning is raised and the transformation
    is not applied.
    """

    def apply(self, model):
        graph_modified = False

        invalid_nodes = _find_invalid_nodes(model)
        if len(invalid_nodes) > 0:
            warnings.warn(
                "Transformation is not applied,\
                 found unsupported nodes in the graph: {}.".format(
                    invalid_nodes
                )
            )
            return (model, graph_modified)

        # Infer the shapes of each tensor, remove unused tensors
        # and give each tensor a readable name
        model = model.transform(InferShapes())
        model = model.transform(RemoveUnusedTensors())

        # This list contains all nodes with initializers that need to be converted
        nodes_with_initializers = ["Mul", "Conv", "Add", "Div", "Reshape"]
        # Obtain a list of initializer names (used to filter out only value infos)
        initializers_names = [x.name for x in model.graph.initializer]

        all_tensors = {}
        # Extract the inputs
        all_tensors = {
            **all_tensors,
            **{x.name: [x.type.tensor_type.elem_type, model.get_tensor_shape(x.name)] for x in model.graph.input},
        }
        # Extract only the output tensors
        all_tensors = {
            **all_tensors,
            **{
                x.name: [x.type.tensor_type.elem_type, model.get_tensor_shape(x.name)]
                for x in model.graph.value_info
                if x.name not in initializers_names
            },
        }
        # Extract only initializers from nodes that are relevant for conversion
        all_tensors = {
            **all_tensors,
            **{
                x.name: [x.data_type, x.dims]
                for x in model.graph.initializer
                if model.find_consumer(x.name).op_type in nodes_with_initializers
            },
        }
        # Extract the outputs
        all_tensors = {
            **all_tensors,
            **{x.name: [x.type.tensor_type.elem_type, model.get_tensor_shape(x.name)] for x in model.graph.output},
        }

        # The list below contains tensor names that are the output of nodes that
        # reduce the tensor's dimension. The shape of these tensors also needs
        # to be extended
        tensors_reduced_dimension = []
        for n in model.graph.node:
            node_op_type = n.op_type
            input_shape = model.get_tensor_shape(n.input[0])
            # Find tensors that are the output of nodes that reduce the dimension
            if node_op_type == "ArgMax":
                keep_dims = get_by_name(n.attribute, "keepdims", "name").i
                if len(input_shape) == 3 and keep_dims == 0:
                    node_out = n.output
                    for n_o in node_out:
                        tensors_reduced_dimension.append(n_o)
            # Each node from the list of supported nodes is made compatible
            # with 4D tensors
            if node_op_type == "Transpose":
                perm = get_by_name(n.attribute, "perm", "name").ints
                if len(perm) == 3:  # Meaning that the transpose operation was on a 3D tensor
                    perm.append(3)  # append 4th dimension
            elif node_op_type in ["ArgMax", "LogSoftMax", "TopK", "Flatten"]:
                axis = get_by_name(n.attribute, "axis", "name")
                if len(input_shape) == 3 and axis.i < 0:
                    axis.i = 3 + axis.i  # count dimensions from the front
            elif node_op_type == "Conv":
                dilations = get_by_name(n.attribute, "dilations", "name").ints
                kernel_shape = get_by_name(n.attribute, "kernel_shape", "name").ints
                pads = get_by_name(n.attribute, "pads", "name").ints
                strides = get_by_name(n.attribute, "strides", "name").ints
                if len(dilations) == 1:  # we must add another dimension to it
                    dilations.append(1)  # only equal dilation value along each spatial axis is supported
                if len(kernel_shape) == 1:  # we must add another dimension to it
                    kernel_shape.append(1)
                if len(pads) == 2:  # pads = [x1_begin, x1_end] --> [x1_begin, x2_begin, x1_end, x2_end]
                    pads.insert(1, 0)
                    pads.append(0)
                if len(strides) == 1:  # strides = [stride_h, stride_w]
                    strides.append(1)
            elif node_op_type == "MaxPool":
                kernel_shape = get_by_name(n.attribute, "kernel_shape", "name").ints
                pads = get_by_name(n.attribute, "pads", "name").ints
                strides = get_by_name(n.attribute, "strides", "name").ints
                if len(kernel_shape) == 1:  # we must add another dimension to it
                    kernel_shape.append(1)
                if len(pads) == 2:  # pads = [x1_begin, x1_end] --> [x1_begin, x2_begin, x1_end, x2_end]
                    pads.insert(1, 0)
                    pads.append(0)
                if len(strides) == 1:  # strides = [stride_h, stride_w]
                    strides.append(1)
            elif node_op_type == "Upsample":
                # the parameter for the Upsample layer is a 1d list of
                # upsampling factors along each axis
                scales = model.get_initializer(n.input[1])
                assert list(scales.shape) == [3]
                scales = np.append(scales, np.asarray(1.0, dtype=np.float32))
                model.set_initializer(n.input[1], scales)
            elif node_op_type == "Resize":
                assert "axes" not in [x.name for x in n.attribute], "%s: Axes attribute is not supported." % n.name
                assert not (len(n.input) in (3, 4) and model.get_initializer(n.input[1]) is not None), (
                    "%s: ROI input is not supported." % n.name
                )
                if len(n.input) == 2:
                    # Resize version 10
                    scales = model.get_initializer(n.input[1])
                    scales = np.append(scales, np.asarray(1.0, dtype=np.float32))
                    model.set_initializer(n.input[1], scales)
                elif len(n.input) == 3:
                    # Resize version 11 and up (no size input)
                    scales = model.get_initializer(n.input[2])
                    scales = np.append(scales, np.asarray(1.0, dtype=np.float32))
                    model.set_initializer(n.input[2], scales)
                elif len(n.input) == 4:
                    scales_exists = (model.get_initializer(n.input[2]) is not None) and (
                        len(model.get_initializer(n.input[2])) != 0
                    )
                    sizes_exists = (model.get_initializer(n.input[3]) is not None) and (
                        len(model.get_initializer(n.input[3])) != 0
                    )
                    assert scales_exists ^ sizes_exists, (
                        "%s: Either scales or the target output size must "
                        "be specified. Specifying both is prohibited." % n.name
                    )
                    if scales_exists:
                        # Scales parameter is a 1d list of upsampling factors along each axis
                        scales = model.get_initializer(n.input[2])
                        scales = np.append(scales, np.asarray(1.0, dtype=np.float32))
                        model.set_initializer(n.input[2], scales)
                    else:
                        # Size parameter is a 1d list of the target size along each axis
                        sizes = model.get_initializer(n.input[3])
                        sizes = np.append(sizes, np.asarray(1.0, dtype=np.int64))
                        model.set_initializer(n.input[3], sizes)
                input_shape.append(1)

        # Change format of each input/value_info/output tensor
        for k, v in all_tensors.items():
            tensor_type = v[0]
            shape = v[1]
            # Add extra dimension for tensors that either:
            # 1) Have 3 dimensions ( (N,C,H) -> (N,C,H,1) )
            # 2) Come after operations that reduce their dimension: e.g. {Argmax, ...}
            if len(shape) == 3 or k in tensors_reduced_dimension:
                shape.append(1)
                model.set_tensor_shape(k, shape, tensor_type)

        return (model, graph_modified)



# Copyright (c) 2023 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of qonnx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
import warnings
from copy import deepcopy
from functools import reduce
from typing import Dict, Tuple

from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.base import Transformation
from qonnx.util.basic import get_by_name

# these ops propagate sparsity masks both forward and backward
# (e.g. a sparse channel on the input creates a sparse channel
# on the output, and vice versa)
# note on Quant and MultiThreshold: only bias-free (no zeropoint)
# versions of these ops are bidirectional
eltwise_ops_bidirectional = [
    "Mul",
    "Div",
    "MultiThreshold",
    "Quant",
    "QuantizeLinear",
    "DequantizeLinear",
    "Clip",
    "Relu",
    "MaxPool",
]

# these ops propagate sparsity masks only backward, not forward
# (e.g. a sparse channel on the output creates a sparse channel
# on the input, but not vice versa)
eltwise_ops_bwdonly = ["Add", "Sub", "BatchNormalization"]

# other ops (MatMul, Conv) have more specialized behavior
# and will be handled by update_node_mask

# mapping of weight tensor axes to input/output channels
optype_to_w_axis = {
    "MatMul": {"in": 0, "out": 1},
    "Conv": {"in": 1, "out": 0},
}


def ensure_masktype_is_dict(mask):
    if isinstance(mask, dict):
        # all good, return as is
        return mask
    if mask is None:
        # use empty dict instead of no sparsity mask (None)
        return dict()
    else:
        raise Exception("Cannot turn %s into dict" % str(mask))


def merge_dicts_of_sets(dict1, dict2):
    ret = deepcopy(dict1)
    for key, val in dict2.items():
        if key in ret.keys():
            ret[key].update(val)
        else:
            ret[key] = val
    return ret


def remove_masked_tensor_channels(tensor_or_shape, mask, axis):
    shape_only = False
    if type(mask) is not list:
        mask_list = list(mask)
    else:
        mask_list = mask
    if type(tensor_or_shape) in [list, tuple]:
        shape_only = True
        tensor_or_shape = np.random.rand(*tensor_or_shape)
    assert type(tensor_or_shape) is np.ndarray
    if tensor_or_shape.ndim == 0 or np.prod(tensor_or_shape.shape) == 1:
        # no pruning for scalar properties
        ret = tensor_or_shape
    else:
        ret = np.delete(tensor_or_shape, mask_list, axis=axis)
    if shape_only:
        return ret.shape
    else:
        return ret


def update_node_mask(node, masks_in, masks_out, lossy=True):
    masks_in = [ensure_masktype_is_dict(x) for x in masks_in]
    masks_out = [ensure_masktype_is_dict(x) for x in masks_out]

    if lossy:
        # when in lossy mode, allow propagation sparsity masks
        # in both directions (e.g. Add nodes will also get pruned)
        update_bidirectional = eltwise_ops_bidirectional + eltwise_ops_bwdonly
        update_bwdonly = []

    if node.op_type in update_bidirectional:
        # any i/o can mask any/all other i/o
        # so just take union
        all_masks = [*masks_in] + [*masks_out]
        ret = reduce(merge_dicts_of_sets, all_masks)
        # duplicate the result for each node input and output
        masks_in = [ret for x in masks_in]
        masks_out = [ret for x in masks_out]
    elif node.op_type in update_bwdonly:
        # output can mask input but not other way around
        all_masks = [*masks_in] + [*masks_out]
        ret = reduce(merge_dicts_of_sets, all_masks)
        masks_in = [ret for x in masks_in]
    elif node.op_type in ["MatMul", "Conv"]:
        # input and output are essentially decoupled from
        # each other by means of the weight (except dwise convs)
        if node.op_type == "Conv":
            groups = get_by_name(node.attribute, "group")
            if groups is not None:
                groups = groups.i
            else:
                groups = 1
            # TODO smarter check, other kinds of grouped convs out there..
            is_depthwise = groups > 1
        else:
            is_depthwise = False

        # convert back to two distinct int sets to be able to use union etc set ops
        w_mask = masks_in[1]
        w_axis_in = optype_to_w_axis[node.op_type]["in"]
        w_axis_out = optype_to_w_axis[node.op_type]["out"]
        w_mask_in = w_mask.get(w_axis_in, set())
        w_mask_out = w_mask.get(w_axis_out, set())
        # take union with i/o masks to update
        conv_io_chan_axis = 1
        i_mask = masks_in[0].get(conv_io_chan_axis, set())
        o_mask = masks_out[0].get(conv_io_chan_axis, set())
        mask_in = w_mask_in.union(i_mask)
        mask_out = w_mask_out.union(o_mask)
        if is_depthwise:
            # depthwise convs couple i<->o channels directly
            mask_in = mask_in.union(mask_out)
            mask_out = mask_in
            # dw convs to only use output side for weights by convention
            w_mask[w_axis_out] = mask_out
        else:
            w_mask[w_axis_out] = mask_out
            w_mask[w_axis_in] = mask_in
        masks_in = [{conv_io_chan_axis: mask_in}, w_mask]
        masks_out = [{conv_io_chan_axis: mask_out}]
    else:
        warnings.warn("Can't propagate sparsity mask through op_type %s" % node.op_type)
    return (masks_in, masks_out)


class ApplyMasks(Transformation):
    """Apply the given sparsity masks in prune_spec to the appropriately named
    tensors in the model. These masks are only annotations, no actual pruning
    is performed at this stage."""

    def __init__(self, prune_spec: Dict) -> None:
        super().__init__()
        self.prune_spec = prune_spec

    def apply(self, model: ModelWrapper) -> Tuple[ModelWrapper, bool]:
        # sanity check:
        # - prune spec must be a dict
        assert isinstance(self.prune_spec, dict)
        for key, val in self.prune_spec.items():
            # sanity check:
            # - prune spec keys must be strings (tensor names)
            assert isinstance(key, str)
            # - prune spec vals must also be dicts
            assert isinstance(val, dict)
            model.set_tensor_sparsity(key, val)
        return (model, False)


class PropagateMasks(Transformation):
    """Propagate the sparsity masks in the network to relevant upstream and
    downstream layers. Some inital sparsity masks must have been applied
    either manually or with the ApplyMasks transformation. Note that not all
    layer types are supported; see the update_node_mask function for details."""

    def __init__(self, lossy: bool = True) -> None:
        super().__init__()
        self.lossy = lossy

    def apply(self, model: ModelWrapper) -> Tuple[ModelWrapper, bool]:
        need_rerun = False
        # traverse graph from inputs to outputs to propagate
        # sparsity masks via per-layer handlers
        for node in model.graph.node:
            node_masks_in = [model.get_tensor_sparsity(x) for x in node.input]
            node_masks_out = [model.get_tensor_sparsity(x) for x in node.output]
            # ensure all mask types are considered as sets
            # otherwise we end up comparing None and dict()
            node_masks_in = [ensure_masktype_is_dict(x) for x in node_masks_in]
            node_masks_out = [ensure_masktype_is_dict(x) for x in node_masks_out]
            (new_in, new_out) = update_node_mask(node, node_masks_in, node_masks_out)
            in_changed = new_in != node_masks_in
            out_changed = new_out != node_masks_out
            need_rerun |= in_changed
            need_rerun |= out_changed
            for inp_name, inp_annot in zip(node.input, new_in):
                model.set_tensor_sparsity(inp_name, inp_annot)
            for out_name, out_annot in zip(node.output, new_out):
                model.set_tensor_sparsity(out_name, out_annot)
        return (model, need_rerun)


class RemoveMaskedChannels(Transformation):
    """Remove channels indicated by sparsity masks on the model. The sparsity
    mask annotations will be removed after they have been processed for each
    tensor. Does not perform any shape consistency checking and may result in
    a broken graph."""

    def __init__(self, lossy: bool = True) -> None:
        super().__init__()
        self.lossy = lossy

    def apply(self, model: ModelWrapper) -> Tuple[ModelWrapper, bool]:
        need_rerun = False
        for node in model.graph.node:
            for ioname in [*node.input, *node.output]:
                io_t = model.get_initializer(ioname)
                io_shp = model.get_tensor_shape(ioname)
                mask = model.get_tensor_sparsity(ioname)
                if mask is None or mask == {}:
                    continue
                # print("[RemoveMaskedChannels] tensor %s mask %s: old shape %s" % (ioname, str(mask), str(io_shp)))
                if io_t is None:
                    # dynamic input/output, no initializer
                    # compute new shape only
                    for target_axis, axis_mask in mask.items():
                        new_shp = remove_masked_tensor_channels(io_shp, axis_mask, axis=target_axis)
                        model.set_tensor_shape(ioname, new_shp)
                else:
                    if node.op_type in ["MatMul"]:
                        w_axis_in = optype_to_w_axis[node.op_type]["in"]
                        w_axis_out = optype_to_w_axis[node.op_type]["out"]
                        w_mask_in = mask.get(w_axis_in, set())
                        w_mask_out = mask.get(w_axis_out, set())
                        new_t = remove_masked_tensor_channels(io_t, w_mask_in, axis=w_axis_in)
                        new_t = remove_masked_tensor_channels(new_t, w_mask_out, axis=w_axis_out)
                        model.set_initializer(ioname, new_t)
                    elif node.op_type in ["Conv"]:
                        w_axis_in = optype_to_w_axis[node.op_type]["in"]
                        w_axis_out = optype_to_w_axis[node.op_type]["out"]
                        w_mask_in = mask.get(w_axis_in, set())
                        w_mask_out = mask.get(w_axis_out, set())
                        ifm_w = io_shp[1]
                        groups = get_by_name(node.attribute, "group")
                        if groups is not None:
                            groups = groups.i
                        else:
                            groups = 1
                        assert groups == 1 or ifm_w == 1, "Unknown grouped conv setting"
                        depthwise = groups > 1
                        if depthwise:
                            # depthwise convs only use the o_mask by convention
                            new_t = remove_masked_tensor_channels(io_t, w_mask_out, axis=w_axis_out)
                            # need to update the group attribute to match new n chans
                            get_by_name(node.attribute, "group").i = new_t.shape[0]
                        else:
                            new_t = remove_masked_tensor_channels(io_t, w_mask_in, axis=w_axis_in)
                            new_t = remove_masked_tensor_channels(new_t, w_mask_out, axis=w_axis_out)
                        model.set_initializer(ioname, new_t)
                    else:
                        new_t = io_t
                        for target_axis, axis_mask in mask.items():
                            if target_axis >= new_t.ndim:
                                # for layers that use broadcasting, param dims have lower dim than input dims
                                # and the target axis won't exist. try to handle those cases appropriately
                                if new_t.ndim == 1:
                                    new_t = remove_masked_tensor_channels(new_t, axis_mask, axis=0)
                                    model.set_initializer(ioname, new_t)
                                elif new_t.ndim == 0:
                                    # don't prune scalar param
                                    continue
                                else:
                                    assert False, "Cannot prune paramet tensor %s with shape %s using spec %s" % (
                                        ioname,
                                        str(io_shp),
                                        str(mask),
                                    )
                            else:
                                new_t = remove_masked_tensor_channels(new_t, axis_mask, axis=target_axis)
                                model.set_initializer(ioname, new_t)
                # clear sparsity annotation since it's already handled
                # leftover annotations here can lead to erronous removal later
                model.set_tensor_sparsity(ioname, {})
                new_shp = model.get_tensor_shape(ioname)
                # print("[RemoveMaskedChannels] tensor %s : new shape %s" % (ioname, str(new_shp)))
                need_rerun = True
        return (model, need_rerun)


class PruneChannels(Transformation):
    """
    Prune channels from specified tensors and their dependencies from a model, as
    specified by the dictionary given in prune_spec.
    This dictionary must be formatted as {tensor_name : {axis : {channels}}}.
    See test_pruning.py for examples.
    If lossy is True, the transformation will aggresively prune all relevant
    upstream/downstream layers around the specified tensors. This is good for
    maintaining the consistency of layer shapes, but may introduce a larger accuracy
    penalty. If lossy is False, the pruning will be more conservative to preserve the
    numerical ranges (e.g. biases won't be pruned in the downstream layers) but this
    may lead to inconsistent shapes in the network.
    """

    def __init__(self, prune_spec: Dict, lossy: bool = True) -> None:
        super().__init__()
        self.prune_spec = prune_spec
        self.lossy = lossy
        if not lossy:
            warnings.warn("The current implementation for lossless channel pruning will fail for some topologies")

    def apply(self, model: ModelWrapper) -> Tuple[ModelWrapper, bool]:
        # check for known patterns that break the pruning transformation
        conv_nodes = model.get_nodes_by_op_type("Conv")
        matmul_nodes = model.get_nodes_by_op_type("MatMul")
        convs_with_bias = [x for x in conv_nodes if len(x.input) == 3]
        if len(convs_with_bias) > 0:
            assert False, "Found Conv nodes with bias, please use cleanup with extract_conv_bias=True first: %s" % str(
                [x.name for x in convs_with_bias]
            )
        dotprod_nodes = conv_nodes + matmul_nodes
        dotprod_nodes_dyn_w = [x for x in dotprod_nodes if model.get_initializer(x.input[1]) is None]
        if len(dotprod_nodes_dyn_w) > 0:
            assert False, (
                "Found MatMul or Conv nodes with non-static weights. "
                "If this is due to weight quantizer ops, try cleanup with preserve_qnt_ops=False: "
                + str([x.name for x in dotprod_nodes_dyn_w])
            )
        model = model.transform(ApplyMasks(self.prune_spec))
        model = model.transform(PropagateMasks(self.lossy))
        model = model.transform(RemoveMaskedChannels(self.lossy))
        return (model, False)



from qonnx.custom_op.channels_last.batch_normalization import BatchNormalization
from qonnx.custom_op.channels_last.conv import Conv
from qonnx.custom_op.channels_last.max_pool import MaxPool

custom_op = dict()

custom_op["Conv"] = Conv
custom_op["MaxPool"] = MaxPool
custom_op["BatchNormalization"] = BatchNormalization



# Copyright (c) 2024 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of qonnx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest

import os
import urllib.request

from qonnx.analysis.inference_cost import aggregate_dict_keys
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.util.cleanup import cleanup
from qonnx.util.inference_cost import inference_cost as infca

download_url = "https://github.com/onnx/models/raw/main/validated/vision/"
download_url += "classification/resnet/model/resnet18-v1-7.onnx?download="

model_details = {
    "resnet18-v1-7": {
        "description": "Resnet18 Opset version 7.",
        "url": download_url,
        "enc": {
            "a": "op_mac_FLOAT32_FLOAT32",
            "b": "total_mem_w_bits",
            "c": "total_mem_w_elems",
            "d": "total_mem_o_bits",
            "e": "total_mem_o_elems",
        },
    },
}


def download_model(test_model, do_cleanup=False, return_modelwrapper=False):
    qonnx_url = model_details[test_model]["url"]
    # download test data
    dl_dir = "/tmp"
    dl_file = dl_dir + f"/{test_model}.onnx"
    ret = dl_file
    if not os.path.isfile(dl_file):
        urllib.request.urlretrieve(qonnx_url, dl_file)
    if do_cleanup:
        out_file = dl_dir + f"/{test_model}_clean.onnx"
        cleanup(dl_file, out_file=out_file, override_inpsize=1)
        ret = out_file
    if return_modelwrapper:
        ret = ModelWrapper(ret)
    return ret


@pytest.mark.parametrize("test_model", model_details.keys())
def test_inference_cost_breakdown(test_model):
    test_details = model_details[test_model]
    model = download_model(test_model, do_cleanup=True, return_modelwrapper=True)
    inf_cost = infca(model, discount_sparsity=False, cost_breakdown=True)
    assert inf_cost["node_cost"]["Conv_0"]["total_macs"] == 118013952
    assert inf_cost["node_cost"]["Conv_1"]["total_macs"] == 115605504
    assert inf_cost["optype_cost"]["Conv"]["total_macs"] == 1813561344
    t_cost = inf_cost["total_cost"]  # total cost
    op_cost = aggregate_dict_keys(inf_cost["optype_cost"])  # cost per optype
    n_cost = aggregate_dict_keys(inf_cost["node_cost"])  # cost per node.
    enc = test_details["enc"]
    assert t_cost[enc["a"]] == op_cost[enc["a"]] == n_cost[enc["a"]], "inf discrepancy"
    assert t_cost[enc["b"]] == op_cost[enc["b"]] == n_cost[enc["b"]], "inf discrepancy"
    assert t_cost[enc["c"]] == op_cost[enc["c"]] == n_cost[enc["c"]], "inf discrepancy"
    assert t_cost[enc["d"]] == op_cost[enc["d"]] == n_cost[enc["d"]], "inf discrepancy"
    assert t_cost[enc["e"]] == op_cost[enc["e"]] == n_cost[enc["e"]], "inf discrepancy"



# Copyright (c) 2023, Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of qonnx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from qonnx.transformation.expose_intermediate import ExposeIntermediateTensorsPatternList
from qonnx.transformation.remove import RemoveUnusedNodes
from qonnx.util.test import download_model


def test_remove_unused_nodes():
    model = download_model("FINN-TFC_W2A2", do_cleanup=True, return_modelwrapper=True)
    orig_output = model.graph.output[0]
    # break out intermediate output
    pattern_list = ["MatMul_0_out0"]
    model = model.transform(ExposeIntermediateTensorsPatternList(pattern_list, dynamic_only=True))
    # remove original top-level output
    model.graph.output.remove(orig_output)
    assert len(model.graph.output) == 1
    assert model.graph.output[0].name == "MatMul_0_out0"
    # call transform to remove the now-dangling tail nodes
    model = model.transform(RemoveUnusedNodes())
    assert len(model.graph.node) == 6



# Copyright (c) 2023 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of qonnx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest

import numpy as np
import onnx
import onnx.parser as oprs
from onnx.helper import make_opsetid

import qonnx.core.onnx_exec as oxe
from qonnx.core.datatype import DataType
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.change_3d_tensors_to_4d import Change3DTo4DTensors
from qonnx.transformation.infer_shapes import InferShapes
from qonnx.util.basic import gen_finn_dt_tensor, qonnx_make_model


def generate_random_input(model):
    """
    Creates input dictionary with a random numpy array
    that matches the input tensor shape.
    """
    input_dict = {}
    for i in range(len(model.graph.input)):
        input_node = model.graph.input[i]
        input_node_name = input_node.name
        input_node_shape = model.get_tensor_shape(input_node_name)
        i_val = gen_finn_dt_tensor(DataType["FLOAT32"], input_node_shape)
        input_dict[input_node_name] = i_val
    return input_dict


def set_all_initializers(model):
    """Sets all initializers of the graph to a random value."""
    for n in model.graph.node:
        if len(n.input) > 1 and n.name != "TopK1":
            init_name = n.input[1]
            init_shape = model.get_tensor_shape(init_name)
            init_val = gen_finn_dt_tensor(DataType["FLOAT32"], init_shape)
            model.set_initializer(init_name, init_val)


def create_arbitrary_model(invalid=False):
    """
    Creates arbitrary model for testing the 3D to 4D transform.
    This model is based on a subpart of QuartzNet.
    """

    Mul1_node = onnx.helper.make_node(
        "Mul",
        inputs=["in1_mul1", "in2_mul1"],  # inputs
        outputs=["out1_mul1"],  # outputs
        name="Mul1",  # name
    )

    Conv1_node = onnx.helper.make_node(
        "Conv",
        inputs=["out1_mul1", "in2_conv1"],
        outputs=["out1_conv1"],
        name="Conv1",
        dilations=[1],
        group=1,
        kernel_shape=[1],
        pads=[0, 0],
        strides=[1],
    )

    if invalid is True:  # To make the graph invalid, a ReLU node is added after the Conv node
        Relu1_node = onnx.helper.make_node("Relu", inputs=["out1_conv1"], outputs=["out1_relu1"], name="Relu1")
        Add1_node = onnx.helper.make_node("Add", inputs=["out1_relu1", "in2_add1"], outputs=["out1_add1"], name="Add1")
    else:
        Add1_node = onnx.helper.make_node("Add", inputs=["out1_conv1", "in2_add1"], outputs=["out1_add1"], name="Add1")

    Mul2_node = onnx.helper.make_node("Mul", inputs=["out1_add1", "in2_mul2"], outputs=["out1_mul2"], name="Mul2")

    Transpose1_node = onnx.helper.make_node(
        "Transpose",
        inputs=["out1_mul2"],
        outputs=["out1_transpose1"],
        name="Transpose1",
        perm=[0, 2, 1],
    )

    LogSoftmax1_node = onnx.helper.make_node(
        "LogSoftmax",
        inputs=["out1_transpose1"],
        outputs=["out1_logsoftmax1"],
        name="LogSoftmax1",
        axis=2,
    )

    ArgMax1_node = onnx.helper.make_node(
        "ArgMax",
        inputs=["out1_logsoftmax1"],
        outputs=["out1_argmax1"],
        name="ArgMax1",
        axis=-1,
        keepdims=0,
    )

    # Inputs and outputs
    in1_mul1 = onnx.helper.make_tensor_value_info("in1_mul1", onnx.TensorProto.FLOAT, [1, 1024, 128])
    out1_argmax1 = onnx.helper.make_tensor_value_info("out1_argmax1", onnx.TensorProto.INT64, [1, 128])

    # Value infos
    out1_mul1 = onnx.helper.make_tensor_value_info("out1_mul1", onnx.TensorProto.FLOAT, [1, 1024, 128])
    out1_conv1 = onnx.helper.make_tensor_value_info("out1_conv1", onnx.TensorProto.FLOAT, [1, 29, 128])

    if invalid is True:
        out1_relu1 = onnx.helper.make_tensor_value_info("out1_relu1", onnx.TensorProto.FLOAT, [1, 29, 128])

    out1_add1 = onnx.helper.make_tensor_value_info("out1_add1", onnx.TensorProto.FLOAT, [1, 29, 128])

    out1_mul2 = onnx.helper.make_tensor_value_info("out1_mul2", onnx.TensorProto.FLOAT, [1, 29, 128])
    out1_transpose1 = onnx.helper.make_tensor_value_info("out1_transpose1", onnx.TensorProto.FLOAT, [1, 128, 29])
    out1_logsoftmax1 = onnx.helper.make_tensor_value_info("out1_logsoftmax1", onnx.TensorProto.FLOAT, [1, 128, 29])

    # Initializers
    in2_mul1 = onnx.helper.make_tensor_value_info("in2_mul1", onnx.TensorProto.FLOAT, [1])
    in2_conv1 = onnx.helper.make_tensor_value_info("in2_conv1", onnx.TensorProto.FLOAT, [29, 1024, 1])
    in2_add1 = onnx.helper.make_tensor_value_info("in2_add1", onnx.TensorProto.FLOAT, [1, 29, 1])
    in2_mul2 = onnx.helper.make_tensor_value_info("in2_mul2", onnx.TensorProto.FLOAT, [1])

    list_of_nodes = [
        Mul1_node,
        Conv1_node,
        Add1_node,
        Mul2_node,
        Transpose1_node,
        LogSoftmax1_node,
        ArgMax1_node,
    ]
    list_of_value_infos = [
        out1_mul1,
        out1_conv1,
        out1_add1,
        out1_mul2,
        out1_transpose1,
        out1_logsoftmax1,
        in2_mul1,
        in2_conv1,
        in2_add1,
        in2_mul2,
    ]

    if invalid is True:
        list_of_nodes.insert(2, Relu1_node)
        list_of_value_infos.append(out1_relu1)

    graph = onnx.helper.make_graph(
        nodes=list_of_nodes,
        name="4d_conversion_test_graph",
        inputs=[in1_mul1],
        outputs=[out1_argmax1],
        value_info=list_of_value_infos,
    )
    onnx_model = qonnx_make_model(graph, producer_name="4d_conversion_test-model")
    model = ModelWrapper(onnx_model)
    set_all_initializers(model)

    return model


def create_arbitrary_model_vgg():
    """
    Creates arbitrary model for testing the 3D to 4D transform.
    This model is based on a subpart of VGG10.
    """
    Conv1_node = onnx.helper.make_node(
        "Conv",
        inputs=["in1_conv1", "in2_conv1"],
        outputs=["out1_conv1"],
        name="Conv1",
        dilations=[1],
        group=1,
        kernel_shape=[3],
        pads=[1, 1],
        strides=[1],
    )

    Div1_node = onnx.helper.make_node("Div", inputs=["out1_conv1", "in2_div1"], outputs=["out1_div1"], name="Div1")

    MaxPool1_node = onnx.helper.make_node(
        "MaxPool",
        inputs=["out1_div1"],
        outputs=["out1_maxpool1"],
        name="MaxPool1",
        kernel_shape=[2],
        pads=[0, 0],
        strides=[2],
    )

    Flatten1_node = onnx.helper.make_node(
        "Flatten",
        inputs=["out1_maxpool1"],
        outputs=["out1_flatten1"],
        name="Flatten1",
        axis=1,
    )

    MatMul1_node = onnx.helper.make_node(
        "MatMul",
        inputs=["out1_flatten1", "in2_matmul1"],
        outputs=["out1_matmul1"],
        name="MatMul1",
    )

    TopK1_node = onnx.helper.make_node(
        "TopK",
        inputs=["out1_matmul1", "in2topk1"],
        outputs=["out1_topk1", "out2_topk1"],
        name="TopK1",
        axis=-1,
        largest=1,
        sorted=1,
    )

    # Inputs and outputs
    in1_conv1 = onnx.helper.make_tensor_value_info("in1_conv1", onnx.TensorProto.FLOAT, [1, 64, 16])
    out2_topk1 = onnx.helper.make_tensor_value_info("out2_topk1", onnx.TensorProto.INT64, [1, 3])

    # Value infos
    out1_conv1 = onnx.helper.make_tensor_value_info("out1_conv1", onnx.TensorProto.FLOAT, [1, 64, 16])
    out1_div1 = onnx.helper.make_tensor_value_info("out1_div1", onnx.TensorProto.FLOAT, [1, 64, 16])
    out1_maxpool1 = onnx.helper.make_tensor_value_info("out1_maxpool1", onnx.TensorProto.FLOAT, [1, 64, 8])
    out1_flatten1 = onnx.helper.make_tensor_value_info("out1_flatten1", onnx.TensorProto.FLOAT, [1, 512])
    out1_matmul1 = onnx.helper.make_tensor_value_info("out1_matmul1", onnx.TensorProto.FLOAT, [1, 24])
    out1_topk1 = onnx.helper.make_tensor_value_info("out1_topk1", onnx.TensorProto.FLOAT, [1, 3])

    # Initializers
    in2_conv1 = onnx.helper.make_tensor_value_info("in2_conv1", onnx.TensorProto.FLOAT, [64, 64, 3])
    in2_div1 = onnx.helper.make_tensor_value_info("in2_div1", onnx.TensorProto.FLOAT, [1])
    in2_matmul1 = onnx.helper.make_tensor_value_info("in2_matmul1", onnx.TensorProto.FLOAT, [512, 24])
    in2topk1 = onnx.helper.make_tensor_value_info("in2topk1", onnx.TensorProto.FLOAT, [1])

    list_of_nodes = [
        Conv1_node,
        Div1_node,
        MaxPool1_node,
        Flatten1_node,
        MatMul1_node,
        TopK1_node,
    ]
    list_of_value_infos = [
        out1_conv1,
        out1_div1,
        out1_maxpool1,
        out1_flatten1,
        out1_matmul1,
        out1_topk1,
        in2_conv1,
        in2_div1,
        in2_matmul1,
        in2topk1,
    ]

    graph = onnx.helper.make_graph(
        nodes=list_of_nodes,
        name="4d_conversion_test_graph",
        inputs=[in1_conv1],
        outputs=[out2_topk1],
        value_info=list_of_value_infos,
    )

    opset_imports = [onnx.helper.make_opsetid("", 11)]
    onnx_model = qonnx_make_model(graph, producer_name="4d_conversion_test-model", opset_imports=opset_imports)
    model = ModelWrapper(onnx_model)

    # Fixed TopK initializer (K=3)
    model.set_initializer("in2topk1", np.array([3]))
    set_all_initializers(model)

    return model


def create_conv_upsample():
    input = """
    <
        ir_version: 7,
        opset_import: ["" : 9]
    >
    agraph (float[1,4,256] in0) => (float[1,2,256] out0)
    <
        float[4,1,16] param_c0_weight,
        float[4] param_c0_bias,
        float[2,4,1] param_c1_weight,
        float[2] param_c1_bias,
        float[3] upsample_scales = {1.0,1.0,2.0}
    >
    {
        c0_out = Conv<dilations=[1],group=4,kernel_shape=[16],pads=[7,7],strides=[2]>(in0, param_c0_weight, param_c0_bias)
        c1_out = Conv<dilations=[1],group=1,kernel_shape=[1],pads=[0,0],strides=[1]>(c0_out, param_c1_weight, param_c1_bias)
        out0 = Upsample<mode="nearest">(c1_out, upsample_scales)
    }
    """
    model = oprs.parse_model(input)
    model = ModelWrapper(model)
    model = model.transform(InferShapes())
    for tensor_name in model.get_all_tensor_names():
        if tensor_name.startswith("param_"):
            init_shape = model.get_tensor_shape(tensor_name)
            model.set_initializer(tensor_name, gen_finn_dt_tensor(DataType["FLOAT32"], init_shape))
    return model


def create_resize(opset):
    """
    Creates an model for testing the 3D to 4D transform of the resize node.
    """
    resize_node1 = onnx.helper.make_node(
        "Resize",
        inputs=["in_resize1", "roi_resize1", "scales_resize1", "sizes_resize1"],
        outputs=["out_resize1"],
        name="Resize1",
        mode="nearest",
    )

    resize_node2 = onnx.helper.make_node(
        "Resize",
        inputs=["out_resize1", "roi_resize2", "scales_resize2"],
        outputs=["out_resize2"],
        name="Resize2",
        mode="nearest",
    )

    in_resize1 = onnx.helper.make_tensor_value_info("in_resize1", onnx.TensorProto.FLOAT, [1, 32, 4])
    out_resize1 = onnx.helper.make_tensor_value_info("out_resize1", onnx.TensorProto.FLOAT, [1, 32, 8])
    out_resize2 = onnx.helper.make_tensor_value_info("out_resize2", onnx.TensorProto.FLOAT, [1, 32, 16])

    roi_resize1 = onnx.helper.make_tensor_value_info("roi_resize1", onnx.TensorProto.FLOAT, [4])
    scales_resize1 = onnx.helper.make_tensor_value_info("scales_resize1", onnx.TensorProto.FLOAT, [])
    sizes_resize1 = onnx.helper.make_tensor_value_info("sizes_resize1", onnx.TensorProto.INT64, [3])

    roi_resize2 = onnx.helper.make_tensor_value_info("roi_resize2", onnx.TensorProto.FLOAT, [4])
    scales_resize2 = onnx.helper.make_tensor_value_info("scales_resize2", onnx.TensorProto.FLOAT, [3])

    list_of_nodes = [
        resize_node1,
        resize_node2,
    ]
    list_of_value_infos = [
        out_resize1,
        roi_resize1,
        sizes_resize1,
        scales_resize1,
        roi_resize2,
        scales_resize2,
    ]

    graph = onnx.helper.make_graph(
        nodes=list_of_nodes,
        name="4d_conversion_resize_test_graph",
        inputs=[in_resize1],
        outputs=[out_resize2],
        value_info=list_of_value_infos,
    )

    onnx_model = qonnx_make_model(
        graph, producer_name="4d_conversion_resize_test-model", opset_imports=[make_opsetid("", opset)]
    )
    model = ModelWrapper(onnx_model)

    model.set_initializer("sizes_resize1", np.array([1, 32, 8], dtype=np.int64))
    if opset == 11:
        model.set_initializer("scales_resize1", np.array([], dtype=np.float32))
    elif opset == 13:
        model.graph.node[0].input[2] = ""
    else:
        assert False, f"Undefined opset {opset} for Resize testcase creator"
    model.set_initializer("scales_resize2", np.array([1.0, 1.0, 2.0], dtype=np.float32))
    model = model.transform(InferShapes())
    return model


@pytest.mark.parametrize("test_model", ["Quartz", "VGG", "ConvUpsample", "Resize11", "Resize13"])
def test_4d_conversion(test_model):
    """
    Test for the 3D to 4D transformation with a valid graph.
    """

    if test_model == "Quartz":
        model = create_arbitrary_model(invalid=False)
    elif test_model == "VGG":
        model = create_arbitrary_model_vgg()
    elif test_model == "ConvUpsample":
        model = create_conv_upsample()
    elif "Resize" in test_model:
        model = create_resize(opset=int(test_model.replace("Resize", "")))
    else:
        raise Exception("Unknown test_model in test_4d_conversion")

    # Inputs
    input_dict = generate_random_input(model)

    # Comparing the outputs of the model before and after the transform
    output_node_name = model.graph.output[0].name
    output_dict = oxe.execute_onnx(model, input_dict, return_full_exec_context=True)
    expected = output_dict[output_node_name]

    model = model.transform(Change3DTo4DTensors())

    for k, v in input_dict.items():
        old_in_name = k
        old_shape = np.shape(v)
        new_in_name = model.graph.input[0].name
        new_shape = old_shape + (1,)
    new_in_val = np.reshape(v, new_shape)
    del input_dict[old_in_name]
    input_dict[new_in_name] = new_in_val

    output_node_name = model.graph.output[0].name
    output_dict = oxe.execute_onnx(model, input_dict, return_full_exec_context=True)
    expected_modified = output_dict[output_node_name]

    expected_modified = np.reshape(expected_modified, np.shape(expected))

    assert (expected == expected_modified).all()


def test_4d_conversion_invalid_nodes():
    """
    Test for the 3D to 4D transformation with an invalid graph.
    """
    model = create_arbitrary_model(invalid=True)

    # Inputs
    input_dict = generate_random_input(model)

    # Initializers
    set_all_initializers(model)

    # Comparing the outputs of the model before and after the transform
    output_node_name = model.graph.output[0].name
    output_dict = oxe.execute_onnx(model, input_dict, return_full_exec_context=True)
    expected = output_dict[output_node_name]

    model = model.transform(Change3DTo4DTensors())

    output_node_name = model.graph.output[0].name
    output_dict = oxe.execute_onnx(model, input_dict, return_full_exec_context=True)
    expected_modified = output_dict[output_node_name]

    expected_modified = np.reshape(expected_modified, np.shape(expected))

    assert (expected == expected_modified).all()



# Copyright (c) 2023 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of qonnx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
from onnx import TensorProto, helper

from qonnx.custom_op.channels_last.base_wrapped_op import ChannelsLastWrappedOp


class BatchNormalization(ChannelsLastWrappedOp):
    def get_nodeattr_types(self):
        """Returns a dict of permitted attributes for node, where:
        ret_dict[attribute_name] = (dtype, require, default_value, <allowed_values>)
        - dtype indicates which member of the ONNX AttributeProto
        will be utilized
        - require indicates whether this attribute is required
        - default_val indicates the default value that will be used if the
        attribute is not set
        - <allowed_values> (if specified) indicates that this attribute can only
        be set to one of the values in the set <allowed_values>. If not specified,
        all values permitted by dtype are allowed.
        """
        return {
            # The epsilon value to use to avoid division by zero.
            "epsilon": ("f", True, 1e-05),
            # Factor used in computing the running mean and variance.
            # e.g., running_mean = running_mean * momentum + mean * (1 - momentum).
            "momentum": ("f", True, 0.9),
        }
        pass

    def make_shape_compatible_op(self, model):
        """Returns a standard ONNX op which is compatible with this CustomOp
        for performing shape inference."""
        # For BatchNorm the output shape should be the same as the input shape.
        # Get the output shape from the input
        out_shape = model.get_tensor_shape(self.onnx_node.input[0])

        # implement tensor with correct shape
        values = np.random.randn(*out_shape).astype(np.float32)
        return helper.make_node(
            "Constant",
            inputs=[],
            outputs=[self.onnx_node.output[0]],
            value=helper.make_tensor(
                name="const_tensor",
                data_type=TensorProto.FLOAT,
                dims=values.shape,
                vals=values.flatten().astype(float),
            ),
            name=self.onnx_node.name,
        )

    def verify_node(self):
        node = self.onnx_node

        verification_successful = True
        info_messages = []

        wrapper_info = ChannelsLastWrappedOp.verify_node(self)
        info_messages.extend(wrapper_info)

        # verify number of attributes
        num_of_attr = 2
        if len(node.attribute) >= num_of_attr:
            info_messages.append("The number of attributes is correct")
        else:
            info_messages.append(
                """The number of attributes is incorrect,
            {} should have {} attributes""".format(
                    node.op_type, num_of_attr
                )
            )
            verification_successful = False

        # verify that all necessary attributes exist
        try:
            self.get_nodeattr("epsilon")
            self.get_nodeattr("momentum")
            info_messages.append("All necessary attributes exist")
        except Exception:
            info_messages.append(
                """The necessary attributes do not exist.
                BatchNormalization needs the following attributes:
                epsilon, momentum"""
            )
            verification_successful = False

        # verify that attributes have the correct datatype.
        try:
            assert isinstance(self.get_nodeattr("epsilon"), float)
            assert isinstance(self.get_nodeattr("momentum"), float)
            info_messages.append("All attributes are of the correct type")
        except Exception:
            info_messages.append("One or more attributes are of the wrong datatype")
            verification_successful = False

        # verify the number of inputs
        if len(node.input) == 5:
            info_messages.append("The number of inputs is correct")
        else:
            info_messages.append("{} needs 5 data inputs".format(node.op_type))
            verification_successful = False

        if not verification_successful:
            raise RuntimeError(
                f"Verification of node {node.name} failed, please check " f"the attached info messages: {info_messages}"
            )

        return info_messages



# Copyright (c) 2020, Xilinx
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of QONNX nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from pkgutil import get_data

import qonnx.core.data_layout as DataLayout
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.fold_constants import FoldConstants
from qonnx.transformation.general import GiveReadableTensorNames, GiveUniqueNodeNames
from qonnx.transformation.infer_data_layouts import InferDataLayouts
from qonnx.transformation.infer_shapes import InferShapes
from qonnx.transformation.lower_convs_to_matmul import LowerConvsToMatMul


def test_infer_data_layouts():
    raw_m = get_data("qonnx.data", "onnx/mnist-conv/model.onnx")
    model = ModelWrapper(raw_m)
    model = model.transform(InferShapes())
    model = model.transform(FoldConstants())
    model = model.transform(GiveUniqueNodeNames())
    model = model.transform(GiveReadableTensorNames())
    model = model.transform(InferDataLayouts())

    assert model.get_tensor_layout("global_in") == DataLayout.NCHW
    assert model.get_tensor_layout("Conv_0_out0") == DataLayout.NCHW
    assert model.get_tensor_layout("MaxPool_0_out0") == DataLayout.NCHW
    assert model.get_tensor_layout("Reshape_0_out0") == DataLayout.NC
    assert model.get_tensor_layout("MatMul_0_out0") == DataLayout.NC
    assert model.get_tensor_layout("global_out") == DataLayout.NC

    model = model.transform(LowerConvsToMatMul())
    model = model.transform(GiveUniqueNodeNames())
    model = model.transform(GiveReadableTensorNames())
    model = model.transform(InferDataLayouts())

    assert model.get_tensor_layout("global_in") == DataLayout.NCHW
    assert model.get_tensor_layout("Transpose_0_out0") == DataLayout.NHWC
    assert model.get_tensor_layout("Im2Col_0_out0") == DataLayout.NHWC
    assert model.get_tensor_layout("MatMul_0_out0") == DataLayout.NHWC
    assert model.get_tensor_layout("MaxPool_0_out0") == DataLayout.NCHW
    assert model.get_tensor_layout("Reshape_0_out0") == DataLayout.NC
    assert model.get_tensor_layout("MatMul_2_out0") == DataLayout.NC
    assert model.get_tensor_layout("global_out") == DataLayout.NC



# Copyright (c) 2020 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
import onnx
import onnx.numpy_helper as np_helper
from pkgutil import get_data

import qonnx.core.onnx_exec as oxe
from qonnx.core.datatype import DataType
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.infer_shapes import InferShapes
from qonnx.util.basic import gen_finn_dt_tensor, qonnx_make_model


def test_mnist_onnx_download_extract_run():
    # load the onnx model
    raw_m = get_data("qonnx.data", "onnx/mnist-conv/model.onnx")
    model = ModelWrapper(raw_m)
    model = model.transform(InferShapes())
    # load one of the test vectors
    raw_i = get_data("qonnx.data", "onnx/mnist-conv/test_data_set_0/input_0.pb")
    raw_o = get_data("qonnx.data", "onnx/mnist-conv/test_data_set_0/output_0.pb")
    input_tensor = onnx.load_tensor_from_string(raw_i)
    output_tensor = onnx.load_tensor_from_string(raw_o)
    # run using QONNX-based execution (full graph)
    input_dict = {"Input3": np_helper.to_array(input_tensor)}
    output_dict = oxe.execute_onnx(model, input_dict, return_full_exec_context=True)
    assert np.isclose(np_helper.to_array(output_tensor), output_dict["Plus214_Output_0"], atol=1e-3).all()
    # test subgraph execution
    start_node = model.graph.node[1]
    end_node = model.graph.node[3]
    subgraph_i_dict = {start_node.input[0]: output_dict[start_node.input[0]]}
    subgraph_o_dict = oxe.execute_onnx(
        model,
        subgraph_i_dict,
        return_full_exec_context=True,
        start_node=start_node,
        end_node=end_node,
    )
    assert np.isclose(subgraph_o_dict[end_node.output[0]], output_dict[end_node.output[0]], atol=1e-3).all()


def test_onnx_exec_internal_rounding():
    inp0 = onnx.helper.make_tensor_value_info("inp0", onnx.TensorProto.FLOAT, [2, 2])
    inp1 = onnx.helper.make_tensor_value_info("inp1", onnx.TensorProto.FLOAT, [1])
    outp = onnx.helper.make_tensor_value_info("outp", onnx.TensorProto.FLOAT, [2, 2])
    mul_node = onnx.helper.make_node("Mul", inputs=["inp0", "inp1"], outputs=["outp"])
    graph = onnx.helper.make_graph(nodes=[mul_node], name="mul_graph", inputs=[inp0, inp1], outputs=[outp])

    model = qonnx_make_model(graph, producer_name="mul-model")
    model = ModelWrapper(model)
    idt = DataType["INT2"]
    model.set_tensor_datatype("inp0", idt)
    model.set_tensor_datatype("inp1", idt)
    model.transform(InferShapes())

    mul_value = np.asarray([-1], dtype=np.float32)
    inp_int = gen_finn_dt_tensor(idt, [2, 2])
    scale = np.random.uniform(low=0, high=1, size=(2, 2)).astype(np.float32)
    inp_rounded = (inp_int * scale) / (scale + 1e-7)
    input_dict = {"inp0": inp_rounded, "inp1": mul_value}
    output_dict = oxe.execute_onnx(model, input_dict)
    produced = output_dict["outp"]
    expected = np.multiply(inp_int, mul_value)
    assert (produced == expected).all()



# Copyright (c) 2020, Xilinx
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of QONNX nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import numpy as np
import warnings

from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.base import Transformation
from qonnx.transformation.infer_shapes import InferShapes
from qonnx.util.basic import get_by_name


class RemoveUnusedNodes(Transformation):
    """Remove nodes which do not contribute to any top-level output in the graph,
    either directly or indirectly."""

    def apply(self, model: ModelWrapper):
        run_again = False
        graph_top_outputs = {x.name for x in model.graph.output}
        for node in model.graph.node:
            successors = model.find_direct_successors(node)
            node_outputs = {x for x in node.output}
            node_top_outputs = graph_top_outputs.intersection(node_outputs)
            if (successors is None) and (len(node_top_outputs) == 0):
                # found node with dangling output, remove
                model.graph.node.remove(node)
                run_again = True
                # remove only one node at a time to avoid potential problems
                # with protobuf container (model.graph.node)
                break

        return (model, run_again)


def remove_node_and_rewire(model, node):
    # Currently cannot remove and rewire join-nodes, probably not necessary to
    # support this
    if model.is_join_node(node):
        # Log this as a warning, so the user is aware of this, there might be
        # somthing wrong or some checks missing at the caller site
        warnings.warn("Removing join-node operation is currently not supported")
        # Exit the function here without doing anything
        return
    # We already know that node is not a join-node, thus to rewire, we only need
    # to check the single producer
    producer = model.find_producer(node.input[0])
    # If there is a producer which is not a fork-node, rewiring is simple
    if producer is not None and not model.is_fork_node(producer):
        # Rewire by skipping the node, letting the producer directly feed the
        # nodes output.
        #   TODO: Check whether this already covers fork-node identities?
        producer.output[0] = node.output[0]
    # If there is no producer or the producer forks, rewiring is a bit more
    # complicated
    else:
        # Now it depends on the successor nodes to rewire their inputs
        successors = model.find_direct_successors(node)
        # Singular node detached from the rest of the graph?
        assert successors is not None, "Whole graph is one node."
        # We need to rewire the input of each successor to not detach parts of
        # the graph
        for successor in successors:
            # Find the inputs of the successor which are produced by the node to
            # be removed
            for i, s_inp in enumerate(successor.input):
                # Note: This might happen multiple times?
                if s_inp == node.output[0]:
                    # Rewire successor's input directly to nodes input
                    # Note: Node may not be a join-node, but there is probably
                    # no such thing as join-node identity anyway
                    successor.input[i] = node.input[0]
    # Remove node
    model.graph.node.remove(node)


class RemoveIdentityOps(Transformation):
    """Remove identity ops like Add/Sub with zero or Mul/Div with one. A tolerance
    value (defaults to 1e-05) can be specified during init for the comparison
    to zero/one."""

    def __init__(self, atol=1e-05):
        super().__init__()
        self.atol = atol

    def apply(self, model):
        opset_version = model.model.opset_import[0].version
        graph = model.graph
        node_ind = 0
        graph_modified = False
        for node in graph.node:
            node_ind += 1
            if node.op_type in ["Add", "Sub"]:
                A = model.get_initializer(node.input[1])
                if A is not None and np.isclose(A, np.zeros_like(A), atol=self.atol).all():
                    remove_node_and_rewire(model, node)
                    graph_modified = True
                    break

            elif node.op_type in ["Mul", "Div"]:
                A = model.get_initializer(node.input[1])
                if A is not None and np.isclose(A, np.ones_like(A), atol=self.atol).all():
                    remove_node_and_rewire(model, node)
                    graph_modified = True
                    break
            elif node.op_type == "Pad":
                pads = get_by_name(node.attribute, "pads")
                if pads is not None:
                    # older versions of Pad op specify pads as attribute
                    pads = np.asarray(pads.ints, dtype=np.int64)
                else:
                    # newer versions of Pad op specify pads as input
                    pads = model.get_initializer(node.input[1])

                if (pads is not None) and (pads == 0).all():
                    remove_node_and_rewire(model, node)
                    graph_modified = True
                    break
            elif node.op_type == "Identity":
                remove_node_and_rewire(model, node)
                graph_modified = True
                break
            elif node.op_type == "Dropout":
                if opset_version < 12:
                    dropout_ratio = get_by_name(node.attribute, "ratio")
                    dropout_id_cond = not (dropout_ratio is None) and dropout_ratio.f == 0
                else:
                    based_on_inplen = len(node.input) == 1
                    based_on_ratio_inp = (not based_on_inplen) and model.get_initializer(node.input[1]) == 0
                    dropout_id_cond = based_on_inplen or based_on_ratio_inp
                if dropout_id_cond:
                    remove_node_and_rewire(model, node)
                    graph_modified = True
                    break

        model = model.transform(InferShapes())
        return (model, graph_modified)



# Copyright (c) 2021, Xilinx
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of QONNX nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import numpy as np
import warnings
from onnx import TensorProto, helper

from qonnx.core.datatype import DataType
from qonnx.transformation.base import Transformation
from qonnx.transformation.remove import RemoveIdentityOps
from qonnx.util.basic import get_by_name


class GemmToMatMul(Transformation):
    """
    Converts Gemm nodes into a MatMul and an Add node.
    This transformation is built to support version 9 of the Gemm node, as
    documented here: https://github.com/onnx/onnx/blob/master/docs/Changelog.md#Gemm-9
    However, earlier and later versions of the node are likely to work as well.
    Explicitly not supported is the optionality of input C in versions >=11 and
    the broadcast attribute of versions <=6.
    """

    def apply(self, model):
        graph = model.graph
        node_ind = 0
        for n in graph.node:
            node_ind += 1
            if n.op_type == "Gemm":
                # Check for correct ONNX version
                model_onnx_version = model.model.opset_import[0].version
                if model_onnx_version != 9:
                    warnings.warn(
                        f"The GemmToMatMul transformation only offers explicit support "
                        f"for version 9 of the Gemm node, but the ONNX version of the "
                        f"supplied model is {model_onnx_version}. "
                        f"Thus the transformation may fail or "
                        f"return incomplete results."
                    )
                running_node_index = node_ind
                # Transpose A?
                transA = get_by_name(n.attribute, "transA")
                if transA is not None and transA.i:
                    # Insert transpose node
                    shape = model.get_tensor_shape(n.input[0])
                    if shape is not None:
                        shape = tuple(reversed(shape))
                    inp_trans_out = helper.make_tensor_value_info(
                        model.make_new_valueinfo_name(),
                        TensorProto.FLOAT,
                        shape,
                    )
                    graph.value_info.append(inp_trans_out)
                    inp_trans_node = helper.make_node("Transpose", [n.input[0]], [inp_trans_out.name])
                    graph.node.insert(running_node_index, inp_trans_node)
                    running_node_index += 1
                    dt = model.get_tensor_datatype(n.input[0])
                    if dt != DataType["FLOAT32"]:
                        model.set_tensor_datatype(inp_trans_out.name, dt)

                    n.input[0] = inp_trans_out.name

                # Transpose B?
                transB = get_by_name(n.attribute, "transB")
                if transB is not None and transB.i:
                    # Insert transpose node
                    shape = model.get_tensor_shape(n.input[1])
                    if shape is not None:
                        shape = tuple(reversed(shape))
                    inp_trans_out = helper.make_tensor_value_info(
                        model.make_new_valueinfo_name(),
                        TensorProto.FLOAT,
                        shape,
                    )
                    graph.value_info.append(inp_trans_out)
                    inp_trans_node = helper.make_node("Transpose", [n.input[1]], [inp_trans_out.name])
                    graph.node.insert(running_node_index, inp_trans_node)
                    running_node_index += 1
                    # Copy over the datatype
                    dt = model.get_tensor_datatype(n.input[1])
                    if dt != DataType["FLOAT32"]:
                        model.set_tensor_datatype(inp_trans_out.name, dt)

                    n.input[1] = inp_trans_out.name

                # Insert MatMul: A * B
                matMul_node = helper.make_node("MatMul", [n.input[0], n.input[1]], [n.output[0]])
                graph.node.insert(running_node_index, matMul_node)
                matMul_node = graph.node[running_node_index]
                running_node_index += 1

                # Insert Mul: (A*B) * alpha
                alpha = get_by_name(n.attribute, "alpha")
                if alpha is None:
                    alpha = np.array(1.0)
                else:
                    alpha = np.array(alpha.f)
                mul_tensor = helper.make_tensor_value_info(
                    model.make_new_valueinfo_name(),
                    TensorProto.FLOAT,
                    None,
                )
                graph.value_info.append(mul_tensor)
                model.set_initializer(mul_tensor.name, alpha)

                A_shape = model.get_tensor_shape(n.input[0])
                B_shape = model.get_tensor_shape(n.input[1])
                if A_shape is not None and B_shape is not None:
                    shape = [A_shape[0], B_shape[1]]
                else:
                    shape = None
                act_mul_tensor = helper.make_tensor_value_info(
                    model.make_new_valueinfo_name(),
                    TensorProto.FLOAT,
                    shape,
                )
                graph.value_info.append(act_mul_tensor)
                mul_node = helper.make_node(
                    "Mul",
                    [act_mul_tensor.name, mul_tensor.name],
                    [n.output[0]],
                )
                graph.node.insert(running_node_index, mul_node)
                mul_node_main_branch = graph.node[running_node_index]
                running_node_index += 1
                matMul_node.output[0] = act_mul_tensor.name

                # Other branch: Insert Mul: beta * C
                beta = get_by_name(n.attribute, "beta")
                if beta is None:
                    beta = np.array(1.0)
                else:
                    beta = np.array(beta.f)
                mul_tensor = helper.make_tensor_value_info(
                    model.make_new_valueinfo_name(),
                    TensorProto.FLOAT,
                    None,
                )
                graph.value_info.append(mul_tensor)
                model.set_initializer(mul_tensor.name, beta)

                C_shape = model.get_tensor_shape(n.input[2])
                act_mul_tensor = helper.make_tensor_value_info(
                    model.make_new_valueinfo_name(),
                    TensorProto.FLOAT,
                    C_shape,
                )
                graph.value_info.append(act_mul_tensor)
                mul_node = helper.make_node(
                    "Mul",
                    [n.input[2], mul_tensor.name],
                    [act_mul_tensor.name],
                )
                graph.node.insert(running_node_index, mul_node)
                running_node_index += 1
                dt = model.get_tensor_datatype(n.input[2])
                if dt != DataType["FLOAT32"]:
                    model.set_tensor_datatype(act_mul_tensor.name, dt)
                n.input[2] = act_mul_tensor.name

                # Insert Add: ((A*B) * alpha) + (beta * C)
                shape = model.get_tensor_shape(mul_node_main_branch.input[0])
                act_add_tensor = helper.make_tensor_value_info(
                    model.make_new_valueinfo_name(),
                    TensorProto.FLOAT,
                    shape,
                )
                graph.value_info.append(act_add_tensor)
                mul_node_main_branch.output[0] = act_add_tensor.name
                add_node = helper.make_node(
                    "Add",
                    [act_add_tensor.name, n.input[2]],
                    [n.output[0]],
                )

                graph.node.insert(running_node_index, add_node)
                running_node_index += 1

                # Delete Gemm node
                graph.node.remove(n)

                # Remove potential unity multiplications from alpha and beta attributes
                model = model.transform(RemoveIdentityOps())

                return model, True

        return model, False



# Copyright (c) 2020, Xilinx
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of QONNX nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.base import Transformation
from qonnx.transformation.general import SortGraph
from qonnx.util.basic import get_by_name


class ExtendPartition(Transformation):
    """Extends GenericPartition type nodes by inserting the graph pointed to by
    the model attribute.
    Argument 0: extend_index
    * List that contains the node indices of the GenericPartition nodes
    """

    def __init__(self, extend_index):
        super().__init__()
        self.extend_index = extend_index

    def apply(self, model):
        graph = model.graph
        graph_modified = False

        partition_nodes_dict = {ind: n for ind, n in enumerate(graph.node) if n.op_type == "GenericPartition"}

        for k, v in partition_nodes_dict.items():
            if k in self.extend_index:
                path_to_model = get_by_name(v.attribute, "model", "name").s.decode("utf-8")
                model_partition = ModelWrapper(path_to_model)

                # Append nodes
                for partition_node in model_partition.graph.node:
                    graph.node.append(partition_node)

                # Append value infos
                partition_valueinfos = [x.name for x in model_partition.graph.value_info]
                for vi_name in partition_valueinfos:
                    vi = model_partition.get_tensor_valueinfo(vi_name)
                    graph.value_info.append(vi)

                # Append initializers
                partition_initializers = [x for x in model_partition.graph.initializer]
                for i in partition_initializers:
                    graph.initializer.append(i)

                # Append tensor annotations, except for the input/output tensors
                # of the partitioned graph, as these will be present in the
                # 'upper' model.
                in_out_names = [x.name for x in model_partition.graph.input]
                in_out_names += [x.name for x in model_partition.graph.output]
                partition_annotations = [
                    x for x in model_partition.graph.quantization_annotation if x.tensor_name not in in_out_names
                ]
                for a in partition_annotations:
                    graph.quantization_annotation.append(a)

                graph.node.remove(v)
                graph_modified = True

        if graph_modified:
            model = model.transform(SortGraph())

        return (model, graph_modified)



import numpy as np
from tf2onnx.late_rewriters import channel_order_rewriters
from tf2onnx.onnx_opset.math import DirectOp, MatMul
from tf2onnx.onnx_opset.nn import BiasAdd, ConvOp

from .quantizers import get_quant_params


def get_qkeras_onnx_handlers(all_quantizers):
    """Returns the handlers for each kind of layer

    Args:
        all_quantizers: All the quantizers of the model in dictionary format *check

    Returns:
        Dictionary containing the handler information for every type of layer
    """
    return {
        "Conv2D": (conv2d_handler, ["Conv2D", all_quantizers]),
        "MatMul": (dense_handler, ["MatMul", all_quantizers]),
        "BiasAdd": (bias_handler, ["BiasAdd", all_quantizers]),
        "Relu": (relu_handler, ["Relu", all_quantizers]),
        "Identity": (identity_handler, ["Identity", all_quantizers]),
    }


def _extract_node_name(onnx_node, keras_quantizers):
    """

    Args:
        onnx_node: The onnx node to get the information from
        keras_quantizers: The dictionary of all the keras quantizers

    """
    onnx_name = onnx_node.name
    keras_names = keras_quantizers.keys()
    for keras_name in keras_names:
        match = "/" + keras_name + "/"
        if match in onnx_name:
            return keras_name
        elif "Identity" in onnx_name:
            onnx_input = onnx_node.input[0]
            keras_input = keras_quantizers[keras_name]["input"]
            if keras_input in onnx_input:
                return keras_name

    return None


def qlayer_handler(ctx, node, name, args):
    all_quantizers = args[0]
    keras_name = _extract_node_name(node, all_quantizers)
    if not keras_name:
        return  # Not found in quantizers, nothing to do
    quantizers = all_quantizers[keras_name]
    if quantizers.get("kernel_quantizer"):
        weights = node.inputs[1].get_tensor_value(as_list=True)
        quant_params = get_quant_params(weights, quantizers["kernel_quantizer"])
        attr = quant_params["attributes"]
        input_nodes = [node.input[1]]
        for key in quant_params["inputs"].keys():
            name = f"{node.name}_kernel_quantizer_{key}"
            np_val = np.asarray(quant_params["inputs"][key])
            ctx.make_const(name, np_val)
            input_nodes.append(name)
        ctx.insert_new_node_on_input(
            node, "Quant", input_nodes, name=node.name + "_kernel_quantizer", **attr, domain="qonnx"
        )

    if quantizers.get("bias_quantizer") and len(node.input) == 3:
        bias = node.inputs[2].get_tensor_value(as_list=True)
        quant_params = get_quant_params(bias, quantizers["bias_quantizer"])
        attr = quant_params["attributes"]
        input_nodes = [node.input[2]]
        for key in quant_params["inputs"].keys():
            name = f"{node.name}_bias_quantizer_{key}"
            np_val = np.asarray(quant_params["inputs"][key])
            ctx.make_const(name, np_val)
            input_nodes.append(name)
        ctx.insert_new_node_on_input(node, "Quant", input_nodes, name=node.name + "_bias_quantizer", **attr, domain="qonnx")

    if quantizers.get("activation"):
        dtypes = [ctx.get_dtype(node.output[0])]
        quant_params = get_quant_params(None, quantizers["activation"])
        attr = quant_params["attributes"]
        input_nodes = [node.output[0]]
        for key in quant_params["inputs"].keys():
            name = f"{node.name}_activation_quantizer_{key}"
            np_val = np.asarray(quant_params["inputs"][key])
            ctx.make_const(name, np_val)
            input_nodes.append(name)
        quant_act_node = ctx.make_node(
            "Quant",
            input_nodes,
            dtypes=dtypes,
            name=node.name + "_activation_quantizer",
            attr=attr,
            domain="qonnx",
        )
        ctx.insert_node_on_output(quant_act_node, node.output[0])
        ctx.set_shape(quant_act_node.output[0], ctx.get_shape(node.output[0]))


def qact_handler(ctx, node, name, args):
    all_quantizers = args[0]
    keras_name = _extract_node_name(node, all_quantizers)
    if not keras_name:
        return  # Not found in quantizers, nothing to do
    quantizers = all_quantizers[keras_name]
    if quantizers.get("activation"):
        dtypes = [ctx.get_dtype(node.output[0])]
        quant_params = get_quant_params(None, quantizers["activation"])
        attr = quant_params["attributes"]
        input_nodes = [node.output[0]]
        for key in quant_params["inputs"].keys():
            name = f"{node.name}_activation_quantizer_{key}"
            np_val = np.asarray(quant_params["inputs"][key])
            ctx.make_const(name, np_val)
            input_nodes.append(name)
        quant_act_node = ctx.make_node(
            "Quant",
            input_nodes,
            dtypes=dtypes,
            name=node.name + "_activation_quantizer",
            attr=attr,
            domain="qonnx",
        )
        ctx.insert_node_on_output(quant_act_node, node.output[0])
        ctx.set_shape(quant_act_node.output[0], ctx.get_shape(node.output[0]))
        channel_order_rewriters._to_channel_first_handler(ctx, quant_act_node)


def conv2d_handler(ctx, node, name, args):
    ConvOp.any_version(11, ctx, node)
    qlayer_handler(ctx, node, name, args)


def dense_handler(ctx, node, name, args):
    MatMul.version_1(ctx, node)
    qlayer_handler(ctx, node, name, args)


def bias_handler(ctx, node, name, args):
    BiasAdd.version_1(ctx, node)
    all_quantizers = args[0]
    keras_name = _extract_node_name(node, all_quantizers)
    if not keras_name:
        return  # Not found in quantizers, nothing to do
    quantizers = all_quantizers[keras_name]

    if quantizers.get("bias_quantizer"):
        bias = node.inputs[1].get_tensor_value(as_list=True)
        quant_params = get_quant_params(bias, quantizers["bias_quantizer"])
        attr = quant_params["attributes"]
        input_nodes = [node.input[1]]
        for key in quant_params["inputs"].keys():
            name = f"{node.name}_bias_quantizer_{key}"
            np_val = np.asarray(quant_params["inputs"][key])
            ctx.make_const(name, np_val)
            input_nodes.append(name)
        ctx.insert_new_node_on_input(node, "Quant", input_nodes, name=node.name + "_bias_quantizer", **attr, domain="qonnx")

    if quantizers.get("activation"):
        # removes node if added earlier
        remove_node_id = node.input[0]
        remove_node = ctx.get_node_by_output(remove_node_id)
        ctx.replace_all_inputs(node.input[0], remove_node.input[0], ops=None)
        dtypes = [ctx.get_dtype(node.output[0])]
        quant_params = get_quant_params(None, quantizers["activation"])
        attr = quant_params["attributes"]
        input_nodes = [node.output[0]]
        for key in quant_params["inputs"].keys():
            name = f"{node.name}_activation_quantizer_{key}"
            np_val = np.asarray(quant_params["inputs"][key])
            ctx.make_const(name, np_val)
            input_nodes.append(name)
        quant_act_node = ctx.make_node(
            "Quant",
            input_nodes,
            dtypes=dtypes,
            name=node.input[0] + "_activation_quantizer",
            attr=attr,
            domain="qonnx",
        )
        ctx.insert_node_on_output(quant_act_node, node.output[0])
        ctx.set_shape(quant_act_node.output[0], ctx.get_shape(node.output[0]))


def relu_handler(ctx, node, name, args):
    DirectOp.version_1(ctx, node)
    qact_handler(ctx, node, name, args)


def identity_handler(ctx, node, name, args):
    DirectOp.version_1(ctx, node)
    qact_handler(ctx, node, name, args)



# Copyright (c) 2020 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import onnx.helper as oh
from onnx import TensorProto

import qonnx.analysis.topology as ta
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.infer_shapes import InferShapes
from qonnx.util.basic import qonnx_make_model


def test_is_linear_linear():
    top_in = oh.make_tensor_value_info("top_in", TensorProto.FLOAT, [2])
    add_param = oh.make_tensor_value_info("add_param", TensorProto.FLOAT, [2])
    mul_param = oh.make_tensor_value_info("mul_param", TensorProto.FLOAT, [2])
    top_out = oh.make_tensor_value_info("top_out", TensorProto.FLOAT, [2])
    modelproto = qonnx_make_model(
        oh.make_graph(
            name="test",
            inputs=[top_in],
            outputs=[top_out],
            value_info=[add_param, mul_param],
            nodes=[
                oh.make_node("Add", ["top_in", "add_param"], ["middle"]),
                oh.make_node("Mul", ["middle", "mul_param"], ["top_out"]),
            ],
        )
    )
    model = ModelWrapper(modelproto)
    model = model.transform(InferShapes())
    ret = model.analysis(ta.is_linear)
    assert ret["is_linear"] is True


def test_is_linear_forked_node_output():
    top_in = oh.make_tensor_value_info("top_in", TensorProto.FLOAT, [2])
    add_param = oh.make_tensor_value_info("add_param", TensorProto.FLOAT, [2])
    mul0_param = oh.make_tensor_value_info("mul0_param", TensorProto.FLOAT, [2])
    mul1_param = oh.make_tensor_value_info("mul1_param", TensorProto.FLOAT, [2])
    mul0_res = oh.make_tensor_value_info("mul0_res", TensorProto.FLOAT, [2])
    mul1_res = oh.make_tensor_value_info("mul1_res", TensorProto.FLOAT, [2])
    top_out = oh.make_tensor_value_info("top_out", TensorProto.FLOAT, [2])
    modelproto = qonnx_make_model(
        oh.make_graph(
            name="test",
            inputs=[top_in],
            outputs=[top_out],
            value_info=[add_param, mul0_param, mul1_param, mul0_res, mul1_res],
            nodes=[
                oh.make_node("Add", ["top_in", "add_param"], ["middle"]),
                oh.make_node("Mul", ["middle", "mul0_param"], ["mul0_res"]),
                oh.make_node("Mul", ["middle", "mul1_param"], ["mul1_res"]),
                oh.make_node("Add", ["mul0_res", "mul1_res"], ["top_out"]),
            ],
        )
    )
    model = ModelWrapper(modelproto)
    model = model.transform(InferShapes())
    ret = model.analysis(ta.is_linear)
    assert ret["is_linear"] is False



# Copyright (c) 2023 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of qonnx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
from onnx import TensorProto, helper

from qonnx.custom_op.channels_last.base_wrapped_op import ChannelsLastWrappedOp
from qonnx.custom_op.general.maxpoolnhwc import compute_pool_output_dim


class MaxPool(ChannelsLastWrappedOp):
    def get_nodeattr_types(self):
        """Returns a dict of permitted attributes for node, where:
        ret_dict[attribute_name] = (dtype, require, default_value, <allowed_values>)
        - dtype indicates which member of the ONNX AttributeProto
        will be utilized
        - require indicates whether this attribute is required
        - default_val indicates the default value that will be used if the
        attribute is not set
        - <allowed_values> (if specified) indicates that this attribute can only
        be set to one of the values in the set <allowed_values>. If not specified,
        all values permitted by dtype are allowed.
        """
        return {
            # stride and shape of MaxPool kernel
            "kernel_shape": ("ints", True, []),
            "strides": ("ints", True, []),
            # amount of padding to be inserted before/after each non-dummy spatial dim
            # i.e. [H_begin, W_begin, H_end, W_end]
            "pads": ("ints", True, [0, 0, 0, 0]),  # default: no padding
        }

    def make_shape_compatible_op(self, model):
        """Returns a standard ONNX op which is compatible with this CustomOp
        for performing shape inference."""
        # Taken from file src/finn/custom_op/general/maxpoolnhwc.py
        # and function: MaxPoolNHWC.make_shape_compatible_op
        node = self.onnx_node
        iname = node.input[0]
        ishape = model.get_tensor_shape(iname)
        kernel_shape = self.get_nodeattr("kernel_shape")
        pads = self.get_nodeattr("pads")
        strides = self.get_nodeattr("strides")
        ndim = len(ishape)
        assert ndim == 3 or ndim == 4, "ChannelsLast MaxPool currently only supports 3D and 4D input tensors."
        ofm_dims = []
        for i in range(ndim - 2):
            if ndim == 3:
                # padding: [begin, end]
                assert pads[0] == pads[1], "Uneven padding not supported"
            elif ndim == 4:
                # padding: [H_begin, W_begin, H_end, W_end]
                assert pads[0] == pads[2], "Uneven padding not supported"
                assert pads[1] == pads[3], "Uneven padding not supported"
            else:
                raise ValueError(
                    f"Inputs of dimensionality ndim={ndim}, are currently not supported for "
                    f"the channels last MaxPool operation."
                )
            ofm_d = compute_pool_output_dim(ishape[i + 1], kernel_shape[i], strides[i], pads[i])
            ofm_dims.append(ofm_d)

        # Get the number of output channels form the input shape
        out_ch = ishape[-1]

        # implement tensor with correct shape
        output_shape = [
            1,
        ]
        for ofm_d in ofm_dims:
            output_shape.append(ofm_d)
        output_shape.append(out_ch)

        # implement tensor with correct shape
        values = np.random.randn(*output_shape).astype(np.float32)
        return helper.make_node(
            "Constant",
            inputs=[],
            outputs=[self.onnx_node.output[0]],
            value=helper.make_tensor(
                name="const_tensor",
                data_type=TensorProto.FLOAT,
                dims=values.shape,
                vals=values.flatten().astype(float),
            ),
            name=self.onnx_node.name,
        )

    def verify_node(self):
        node = self.onnx_node

        verification_successful = True
        info_messages = []

        wrapper_info = ChannelsLastWrappedOp.verify_node(self)
        info_messages.extend(wrapper_info)

        # verify number of attributes
        num_of_attr_min = 3
        num_of_attr_max = 7
        if (len(node.attribute) >= num_of_attr_min) and len(node.attribute) <= num_of_attr_max:
            info_messages.append("The number of attributes is correct")
        else:
            info_messages.append(
                """The number of attributes is incorrect,
            {} should have between {} and {} attributes""".format(
                    node.op_type, num_of_attr_min, num_of_attr_max
                )
            )
            verification_successful = False

        # verify that all necessary attributes exist
        try:
            self.get_nodeattr("kernel_shape")
            self.get_nodeattr("pads")
            self.get_nodeattr("strides")
            info_messages.append("All necessary attributes exist")
        except Exception:
            info_messages.append(
                """The necessary attributes do not exist.
                MaxPool needs the following attributes:
                kernel_shape, pads, strides"""
            )
            verification_successful = False

        # verify that attributes have the correct datatype.
        try:
            assert isinstance(self.get_nodeattr("kernel_shape"), list)
            assert isinstance(self.get_nodeattr("pads"), list)
            assert isinstance(self.get_nodeattr("strides"), list)
            info_messages.append("All attributes are of the correct type")
        except Exception:
            info_messages.append("One or more attributes are of the wrong datatype")
            verification_successful = False

        # verify the number of inputs
        if len(node.input) == 1:
            info_messages.append("The number of inputs is correct")
        else:
            info_messages.append("{} needs 1 data input".format(node.op_type))
            verification_successful = False

        if not verification_successful:
            raise RuntimeError(
                f"Verification of node {node.name} failed, please check the " f"attached info messages: {info_messages}"
            )

        return info_messages
